"""
The following routine converts a single recipe tree
(:py:class:`~recipe_grid.recipe.RecipeTreeNode`) into the equivalent
:py:class:`~Table` form. It should be used to convert each recipe tree in
:py:attr:`Recipe.recipe_trees <recipe_grid.recipe.Recipe.recipe_trees>` into its own table in the
rendered output.

.. autofunction:: recipe_tree_to_table(recipe_tree: recipe_grid.recipe.RecipeTreeNode) -> recipe_grid.renderer.table.Table

"""  # noqa: E501

from typing import cast

from recipe_grid.recipe import (
    RecipeTreeNode,
    Ingredient,
    Reference,
    Step,
    SubRecipe,
)

from recipe_grid.renderer.table import (
    BorderType,
    Cell,
    Table,
    right_pad_table,
    combine_tables,
    set_border_around_table,
)


def recipe_tree_to_table(
    recipe_tree: RecipeTreeNode, _root: bool = True
) -> Table[RecipeTreeNode]:
    """
    Convert a recipe tree into tabular form.

    To display the resulting :py:class:`~Table`, cell contents should be
    rendered as indicated below.

    Firstly, the obvious cases:

    * :py:class:`~Ingredient` is shown as the ingredient quantity and name etc.
    * :py:class:`~Reference` is shown as its label (and quantity/proportion
      etc.)
    * :py:class:`~Step` is shown as its description. The cells immediately to
      the left in the generated table will contain the inputs to this step.

    Finally, cells containing a :py:class:`~SubRecipe` may appear in the table
    for one of two purposes:

    * For single-output sub recipes, the cell containing the
      :py:class:`SubRecipe` will be located above the cells containing the body
      of the sub recipe. This cell should be rendered in the style of a heading
      and containing the output name as the text.
    * For multiple-output sub recipes, thec cell containing the
      :py:class:`SubRecipe` will be immediately to the right of the cells
      defining the sub-recipe and should be rendered as a list all of the
      output names for the sub recipe, for example as a bulleted list.

      .. note::

          This cell will have all but its left border style set to be
          :py:attr:`BorderType.none` to give the appearance of the output list
          being located out to the right of the rest of the table.

    Unless otherwise stated, all cells will have all borders set to
    :py:attr:`BorderType.normal` with the exception of those borders of cells
    near the edges of a sub recipe block. These will have a style of
    :py:attr:`BorderType.sub_recipe`.

    .. note::

        The ``_root`` argument is for internal use and must be left unspecified
        when called by external users. Internally, it indicates if the node
        passed to this function is the root node of its recipe tree.
    """
    if isinstance(recipe_tree, (Ingredient, Reference)):
        table: Table[RecipeTreeNode] = Table([[Cell(recipe_tree)]])
        if _root:
            return set_border_around_table(table, BorderType.sub_recipe)
        else:
            return table

    elif isinstance(recipe_tree, Step):
        input_tables = [
            recipe_tree_to_table(input_tree, False) for input_tree in recipe_tree.inputs
        ]

        # Pad all input tables to same width and stack them up
        input_columns = max(table.columns for table in input_tables)
        input_tables = [right_pad_table(table, input_columns) for table in input_tables]
        combined_input_tables = combine_tables(input_tables, axis=0)

        # Add step to RHS of inputs
        table = combine_tables(
            [
                combined_input_tables,
                Table.from_dict(
                    {(0, 0): Cell(recipe_tree, rows=combined_input_tables.rows)}
                ),
            ],
            axis=1,
        )

        if _root:
            return set_border_around_table(table, BorderType.sub_recipe)
        else:
            return table
    elif isinstance(recipe_tree, SubRecipe):
        sub_tree_table = recipe_tree_to_table(recipe_tree.sub_tree, False)

        if len(recipe_tree.output_names) == 1:
            if recipe_tree.show_output_names:
                return set_border_around_table(
                    combine_tables(
                        [
                            Table.from_dict(
                                {
                                    (0, 0): Cell(
                                        cast(RecipeTreeNode, recipe_tree),
                                        columns=sub_tree_table.columns,
                                    ),
                                }
                            ),
                            sub_tree_table,
                        ],
                        axis=0,
                    ),
                    BorderType.sub_recipe,
                )
            else:
                return set_border_around_table(sub_tree_table, BorderType.sub_recipe)
        else:
            return combine_tables(
                [
                    set_border_around_table(sub_tree_table, BorderType.sub_recipe),
                    Table.from_dict(
                        {
                            (0, 0): Cell(
                                cast(RecipeTreeNode, recipe_tree),
                                rows=sub_tree_table.rows,
                                border_top=BorderType.none,
                                border_right=BorderType.none,
                                border_bottom=BorderType.none,
                            ),
                        }
                    ),
                ],
                axis=1,
            )
    else:
        raise NotImplementedError(type(recipe_tree))



"""
A Sphinx extension for embedding compiled recipes in HTML outputs.

This extension is enabled by adding ``"recipe_grid.sphinx_ext"`` to your
Sphinx extensions list.

.. highlight:: rst

This extension adds a ``recipe`` directive which may be used to insert recipe
grid tables into (HTML-only) sphinx documentation. For example::

    .. recipe::
        1 can of spam, sliced
        1 egg, boiled

        fried spam and eggs = fry(spam, egg)

Produces the following output:

.. recipe::
    1 can of spam, sliced
    1 egg, boiled

    fried spam and eggs = fry(spam, egg)

Referencing earlier recipes
===========================

Recipes may be defined across several blocks with later blocks referencing
outputs from earlier blocks. For example::

    .. recipe::
        serve(
            fried spam and eggs,
            more spam,
        )

.. recipe::
    serve(
        fried spam and eggs,
        more spam,
    )

Starting a new recipe
=====================

A new, independent recipe may be started by adding the ``:start-new-recipe:``
option enabling ingredient/output names to be reused without referring to
earlier recipes. For example::

    .. recipe::
        :start-new-recipe:

        1 egg, scrambled

.. recipe::
    :start-new-recipe:

    1 egg, scrambled

Scaling
=======

Recipes can be scaled using the ``:scale:`` option::

    .. recipe::
        :scale: 3/2

        1 packet of cupcake batter
        bake(
            pre-heat oven,
            put into {4} cupcake cases(cupcake batter)
        )

.. recipe::
    :scale: 3/2

    1 packet of cupcake batter
    bake(
        pre-heat oven,
        put into {4} cupcake cases(cupcake batter)
    )


.. warning::
    The ``:scale:`` option only scales the recipe shown for the given
    ``recipe`` directive. Where a recipe is given in several blocks, the
    ``:scale:`` option must be repeated otherwise the displayed values will be
    inconsistent.


Showing sources
===============

To facilitate writing of documentation examples for the recipe grid syntax, the
``:show-source:`` option may be used which will cause the rendered recipe grid
table to be shown adjacent to the recipe grid source which defines it. For
example::

    .. recipe::
        :show-source:

        1 oven pizza
        slice(
            bake(
                pre-heat oven,
                oven pizza,
            )
        )

Will be rendered as:

.. recipe::
    :show-source:

    1 oven pizza
    slice(
        bake(
            pre-heat oven,
            oven pizza,
        )
    )

Internals
=========

This Sphinx extension runs as follows:

1. The ``builder-inited`` Sphinx event is used to add the recipe grid CSS to the
   site.
2. For each page in the documentation, recipes are compiled as follows:
    a. ``recipe`` directives turn into ``pending_recipe_node`` nodes and
       capture the recipe grid source in ``env.recipe_sources`` (see
       :py:meth:`RecipeDirective.run`).
    b. In the ``doctree-read`` Sphinx event (called after each page has been
       read and the directives expanded) compiles the recipes into recipe
       grid abstract recipe descriptions (see :py:mod`recipe_grid.recipe`) and
       stores these in the ``pending_recipe_node`` objects.
3. During HTML rendering, ``pending_recipe_node`` nodes are replaced with
   rendered recipe tables.

"""

from typing import Union, List, Mapping, Any

import os

from collections import defaultdict

from fractions import Fraction

from docutils import nodes
from docutils.parsers.rst.directives import flag

from sphinx.application import Sphinx
from sphinx.errors import SphinxError
from sphinx.util.docutils import SphinxDirective
from sphinx.writers.html import HTMLTranslator

from peggie import ParseError

from recipe_grid import __version__
from recipe_grid.compiler import compile, RecipeCompileError
from recipe_grid.recipe import Recipe
from recipe_grid.renderer.html import t, render_recipe_tree
from recipe_grid.static_site.templates import tables_only_css_template
from recipe_grid.number_parser import number


class SphinxRecipeCompileError(SphinxError):
    """Thrown when a recipe fails to compile."""


class RecipeDirective(SphinxDirective):

    has_content = True

    option_spec = {
        "scale": number,
        "show-source": flag,
        "start-new-recipe": flag,
    }

    def run(self) -> List[nodes.Node]:
        # The recipes in a doctree are accumulated in env.recipe_sources in the
        # form:
        #
        #    {docname: [[(source, node), ...], ...], ...}
        #
        # Where the inner most lists contain a complete recipe's complete set
        # of source blocks along with the associated pending_recipe_node
        # placeholder object. The 'source' strings will be padded with leading
        # newlines so that line numbers match their positions in the RST
        # document.
        if not hasattr(self.env, "recipe_sources"):
            self.env.recipe_sources = defaultdict(list)  # type: ignore

        page_recipes = self.env.recipe_sources[self.env.docname]  # type: ignore

        if len(page_recipes) == 0 or "start-new-recipe" in self.options:
            page_recipes.append([])

        # Extract the recipe and replace with a pending_recipe_node
        page_recipe_blocks = page_recipes[-1]
        source = "\n".join(self.content)
        padded_source = "\n" * self.content_offset + source
        node = pending_recipe_node()
        node.scale = self.options.get("scale", 1)
        page_recipe_blocks.append((padded_source, node))

        if "show-source" in self.options:
            return [
                nodes.literal_block(text=source, language="text"),
                node,
            ]
        else:
            return [node]


def compile_recipes(app: Sphinx, doctree: nodes.Node) -> None:
    """Parse and compile the recipes appearing on a page."""
    env = app.env

    # Skip if no recipe sources found anywhere
    if not hasattr(env, "recipe_sources"):
        return

    docname = env.docname

    page_recipes = env.recipe_sources[docname]  # type: ignore

    for i, page_recipe_blocks in enumerate(page_recipes):
        sources = [source for source, _node in page_recipe_blocks]
        try:
            recipes = compile(sources)
        except (ParseError, RecipeCompileError) as e:
            filename = env.doc2path(docname)
            raise SphinxRecipeCompileError(f"Recipe compile error in {filename}: {e}")

        for (_source, node), recipe in zip(page_recipe_blocks, recipes):
            node.id_prefix = f"sub-recipe-{i}-"
            node.recipe = recipe


class pending_recipe_node(nodes.Element):  # type: ignore
    # Set when the node is created by RecipeDirective.run
    scale: Union[int, float, Fraction]

    # The following fields are assigned by compile_recipes
    id_prefix: str
    recipe: Recipe


def visit_pending_recipe_node(self: HTMLTranslator, node: nodes.Node) -> None:
    self.body.append(
        t(
            "div",
            "\n".join(
                render_recipe_tree(recipe_tree, node.id_prefix)
                for recipe_tree in node.recipe.scale(node.scale).recipe_trees
            ),
            class_="rg-recipe-block",
        )
    )


def depart_pending_recipe_node(self: HTMLTranslator, node: nodes.Node) -> None:
    pass


def copy_stylesheet(app: Sphinx) -> None:
    if app.builder.format == "html":
        static_dir = os.path.join(app.builder.outdir, "_static")
        os.makedirs(static_dir, exist_ok=True)
        filename = "recipe_grid_tables.css"
        with open(os.path.join(static_dir, filename), "w") as f:
            f.write(tables_only_css_template.render())
        app.add_css_file(filename)


def setup(app: Sphinx) -> Mapping[str, Any]:
    app.connect("builder-inited", copy_stylesheet)

    app.add_directive("recipe", RecipeDirective)

    app.connect("doctree-read", compile_recipes)

    app.add_node(
        pending_recipe_node,
        html=(visit_pending_recipe_node, depart_pending_recipe_node),
    )

    return {
        "version": __version__,
        "parallel_read_safe": True,
        "parallel_write_safe": True,
    }



"""
Generates a stand alone HTML page for a single recipe.
"""

from typing import Union, Optional

from fractions import Fraction

from functools import partial

from pathlib import Path

# NB: Used in favour of recipe_grid.markdown.compile_markdown due to unified
# exceptions.
from recipe_grid.static_site.recipe_directory import compile_recipe_markdown

from recipe_grid.static_site.templates import standalone_recipe_template

from recipe_grid.static_site.html_postprocessing import (
    postprocess_html,
    embed_local_links_as_data_urls,
)


def generate_standalone_page(
    input_file: Path,
    scale: Optional[Union[int, float, Fraction]] = None,
    servings: Optional[int] = None,
    embed_local_links: bool = True,
) -> str:
    """
    Generate a standalone page with a rendered markdown recipe.

    Parameters
    ==========
    input_file : Path
        The file containing the recipe grid markdown file to compile.
    scale : number or None
        If given, scales the recipe by the provided factor (e.g. scale=2 will
        double all quantities). Must not be given if 'servings' is used.
    servings : int or None
        If given, scales the recipe to the specified number of servings. Will
        fail if the recipe does not specify the number of servings it makes in
        its title. Must not be given if 'scale' is used.
    embed_local_links : bool
        If True, links to local files will be replaced by ``data:`` URLs with
        the referenced file contents embedded.
    """
    recipe = compile_recipe_markdown(
        input_file,
        require_title=False,
        require_servings=servings is not None,
    )

    if scale is not None:
        assert servings is None
    elif servings is not None:
        scale = Fraction(servings, recipe.servings)
    else:
        scale = 1

    recipe_html = recipe.render(scale)

    if embed_local_links:
        recipe_html = postprocess_html(
            recipe_html,
            complete_document=True,
            stages=[
                partial(
                    embed_local_links_as_data_urls,
                    source=input_file,
                    root=input_file.parent,
                ),
            ],
        )

    return standalone_recipe_template.render(
        title=recipe.title if recipe.title is not None else "Recipe",
        body=recipe_html,
    )



r"""
Recipe Grid extends Markdown syntax to provide a convenient way of describing
recipes within a document. See the :ref:`markdown_reference` for an overview of
these extensions.

A recipe markdown document is compiled using:

.. autofunction:: compile_markdown

This produces a :py:class:`MarkdownRecipe` object which in turn may be compiled
into a final HTML form using its :py:meth:`MarkdownRecipe.render` method.

.. autoclass:: MarkdownRecipe
    :members:


.. note::

    Unfortunately, though internally Recipe Grid's Markdown parser is built as
    a :py:mod:`marko` extension, due to limitations of the :py:mod:`marko`
    extension API, this extension is not readily composible with other
    extensions. As a result, this extension is not exposed as a public API.

"""

from typing import Optional, List, MutableMapping, Union, Any, NamedTuple, Match, cast

from marko import Markdown, block, inline, helpers  # type: ignore

import re

import random

import string

import html

from fractions import Fraction

from dataclasses import dataclass, field

from collections import OrderedDict

from peggie.error_message_generation import offset_to_line_and_column

from recipe_grid.recipe import Recipe
from recipe_grid.compiler import compile
from recipe_grid.scaled_value_string import ScaledValueString as SVS
from recipe_grid.renderer.html import (
    render_scaled_value_string,
    render_number,
    render_recipe_tree,
    t,
)


def generate_placeholder(num_random_chars: int = 32) -> str:
    """
    Generate a long random ASCII string for use as a unique placeholder
    to temporarily insert into markdown HTML output and later substitute for a
    compiled recipe table.
    """
    slug = "".join(
        random.choice(string.ascii_uppercase) for _ in range(num_random_chars)
    )
    return f"%{slug}%"


class ScaledValueExpression(inline.InlineElement):  # type: ignore
    """
    Marko syntax extension: adds the recipe grid curly-bracket based scalable
    value syntax.
    """

    # Matches fractions, e.g. "1/2" and "3 1/2"
    fraction_pattern = re.compile(
        r"(?:((?P<integer>[0-9]+)[ \t]+)?(?P<numerator>[0-9]+)[ \t]*/[ \t]*(?P<denominator>[0-9]+))"
    )

    # Matches decimal values, e.g. "123" or "1.234"
    decimal_pattern = re.compile(r"(?P<decimal>[0-9]+(\.[0-9]*)?)")

    # Matches a single non-numerical character or backslash escaped character
    free_text_pattern = re.compile(r"\\(?P<escaped_char>.)|(?P<char>[^0-9\{\}])")

    # Matches either a single fraction, decimal or character.
    any_part_pattern = re.compile(
        r"(?:"
        + fraction_pattern.pattern
        + r"|"
        + decimal_pattern.pattern
        + r"|"
        + free_text_pattern.pattern
        + r")"
    )

    # Matches a curly-bracket enclosed string containing fractions and decimals
    # (which should be scaled) and characters (which should be passed through
    # as-is).
    pattern = re.compile(r"\{(?P<source>" + any_part_pattern.pattern + r"*)\}")

    priority = 6

    string: SVS
    """
    The parsed string within the curly brackets as a
    :py;class:`~ScaledValueString`.
    """

    def __init__(self, match: Match[str]) -> None:
        self.string = SVS(
            [
                # Fraction case
                (
                    (int(submatch["integer"]) if submatch["integer"] is not None else 0)
                    + Fraction(
                        int(submatch["numerator"]),
                        int(submatch["denominator"]),
                    )
                )
                if submatch["numerator"] is not None
                else
                # Decimal case
                (
                    int(float(submatch["decimal"]))
                    if "." not in submatch["decimal"]
                    else float(submatch["decimal"])
                )
                if submatch["decimal"] is not None
                else
                # Escaped char case
                submatch["escaped_char"]
                if submatch["escaped_char"] is not None
                else submatch["char"]
                for submatch in self.any_part_pattern.finditer(match["source"])
            ]
        )


class LogPosMixin:
    """
    A mixin for :py:mod:`marko.block` classes which logs the source offset in
    the :py:attr:`pos` attribute when parsing commences ('pos' in marko
    terminology).
    """

    pos: int

    def __init__(self, *args: Any, pos: int, **kwargs: Any) -> None:
        super().__init__(*args, **kwargs)  # type: ignore
        self.pos = pos

    @classmethod
    def parse(cls, source: helpers.Source) -> "LogPosMixin":
        pos = source.pos
        code = super().parse(source)  # type: ignore
        return cls(code, pos=pos)


class CodeBlock(LogPosMixin, block.CodeBlock):  # type: ignore
    """Marko extension: Adds 'pos' attribute to CodeBlock elements."""

    override = True


class FencedCode(LogPosMixin, block.FencedCode):  # type: ignore
    """Marko extension: Adds 'pos' attribute to FencedCode elements."""

    override = True


class Document(block.Document):  # type: ignore
    """
    Marko extension: Makes a copy of the original markdown source in the
    :py:attr:`text` attribute.
    """

    override = True

    text: str
    """
    A copy of the original markdown source provided to Marko.
    """

    def __init__(self, text: str) -> None:
        super().__init__(text)
        self.text = text


class RecipeSourceBlock(NamedTuple):
    """
    A named tuple containing details of a recipe-containing source block within
    a Markdown document.
    """

    source: str
    """The recipe listing."""

    pos: int
    """The offset of the start of the recipe listing in the provided markdown."""

    in_fenced_block: bool
    """True if found in fenced code block, False if in an indented block."""

    start_of_new_recipe: bool
    """
    Does this source block mark the beginning of a new, independent recipe
    (with its own separate namespace)?
    """

    def get_line_number_corrected_source(self, markdown_source: str) -> str:
        """
        Return a copy of :py:attr:`source` with newlines prepended such that
        line numbers in the source match the corresponding line numbers in the
        markdown document.

        Requires the complete original markdown source listing (as used during
        parsing) as an argument.
        """
        newlines = "\n" * (offset_to_line_and_column(markdown_source, self.pos)[0] - 1)

        # NB: the 'pos' is the offset of the fence, not the first line of
        # fenced code.
        if self.in_fenced_block:
            newlines += "\n"

        return newlines + self.source


@dataclass
class MarkdownRecipe:
    """
    The result of parsing a recipe grid markdown file.

    Most users should refer directly to the :py:meth:`render` method which will
    produce a complete HTML listing for the recipe, optionally after scaling by
    a given factor. Other metadata (e.g. :py:attr:`title` and
    :py:attr:`servings`) may also be useful for indexing purposes.
    """

    html: str = ""
    """
    The compiled HTML for the markdown document with placeholders for various
    recipe-related values.

    Specifically the following placeholders should be replaced with suitable
    HTML to produce a useful web page:

    * The :py:class:`~ScaledValueString` objects in
      :py:attr:`scaled_value_strings` must be scaled, rendered and substituted
      for the placeholders given in that dictionary. These include places where
      the curly-bracket syntax has been used within a recipe, and the title
      itself when it contains a number of servings.

    * The recipes in :py:attr:`recipe_placeholders` must be scaled and rendered
      as HTML tables and inserted in place of the provided placeholders.

    * When given :py:attr:`pre_title_placeholder` and
      :py:attr:`post_title_placeholder` must be replaced with a suitable
      surroundings for the title-containing H1 tag. For example, a ``<header>``
      tag and the addition of notes about how the recipe was scaled.

    All of the above steps are automated by the :py:meth:`render` method.
    """

    title: Optional[str] = None
    """
    The title of the recipe (with the serving count and prepositions removed,
    if present).

    For example, given the markdown title "# Spam for 2", this field will
    become "Spam".

    If None, no (H1 level) title was given in the recipe.
    """

    servings: Optional[int] = None
    """
    The number of servings this recipe is for, if known from the title.
    """

    pre_title_placeholder: Optional[str] = None
    post_title_placeholder: Optional[str] = None
    """
    Strings which may be substituted with content to insert before and after
    the title-containing <h1>, if one was identified successfully.
    """

    scaled_value_strings: MutableMapping[str, SVS] = field(default_factory=OrderedDict)
    """
    Scaled value strings to be substituted within the recipe text. A mapping
    from placeholder string to corresponding :py:class:`~ScaledValueString`.
    """

    recipe_placeholders: MutableMapping[str, Recipe] = field(
        default_factory=OrderedDict
    )
    """
    Compiled recipes, one per recipe code block, alongside the placeholder
    string.
    """

    @property
    def recipes(self) -> List[List[Recipe]]:
        """
        An convenience view of :py;attr:`recipe_placeholders` containing just
        the recipes without placeholder strings.
        """
        out: List[List[Recipe]] = []
        for recipe in self.recipe_placeholders.values():
            if recipe.follows is None:
                out.append([])
            out[-1].append(recipe)
        return out

    def render(self, scale: Union[int, float, Fraction] = 1) -> str:
        """
        Render this recipe scaled by a given factor.
        """
        html = self.html

        # Substitute scaled value strings
        for placeholder, svs in self.scaled_value_strings.items():
            html = html.replace(
                placeholder,
                render_scaled_value_string(svs.scale(scale)),
            )

        # Substitute scaled recipes
        id_prefix_index = 0
        for placeholder, recipe in self.recipe_placeholders.items():
            if recipe.follows is None:
                id_prefix_index += 1

            if id_prefix_index > 1:
                id_prefix = f"recipe{id_prefix_index}-"
            else:
                id_prefix = "recipe-"

            html = html.replace(
                placeholder,
                t(
                    "div",
                    "\n".join(
                        render_recipe_tree(recipe_tree, id_prefix)
                        for recipe_tree in recipe.scale(scale).recipe_trees
                    ),
                    class_="rg-recipe-block",
                ),
            )

        # Modify heading to include scaling info
        if (
            self.title is not None
            and self.pre_title_placeholder is not None
            and self.post_title_placeholder is not None
        ):
            html = html.replace(self.pre_title_placeholder, "<header>")
            post_title_text = ""
            if scale != 1:
                if self.servings is not None:
                    orig_servings = t(
                        "span",
                        f"{self.servings} serving{'s' if self.servings != 1 else ''}",
                        class_="rg-original-servings",
                    )
                    post_title_text = t("p", f"Rescaled from {orig_servings}.")
                else:
                    scale_str = t(
                        "span",
                        f"{render_number(scale)}&times;",
                        class_="rg-scaling-factor",
                    )
                    post_title_text = t("p", f"Scaled {scale_str}")
            html = html.replace(
                self.post_title_placeholder, f"{post_title_text}</header>"
            )

        return html


class RecipeGridRendererMixin:
    """
    Mixin for :py:class:`marko.renderer.Renderer` which collects recipe code
    blocks, and the input markdown source, during rendering.

    Indented code blocks and code blocks fenced code blocks with language
    ``recipe`` or ``new-recipe`` are treated as recipe blocks and extracted.
    All other blocks are treated as ordinary code blocks and left to the
    renderer.
    """

    title_serving_count_pattern = re.compile(
        (
            r"(?P<preposition>((to\s+)?serves?|for|makes|serving)\s+)"
            r"(?P<servings>[0-9]+)\s*"
            r"$"
        ),
        re.IGNORECASE,
    )
    """
    A regex matching recipe a preposition and serving count (e.g. in "Spam for
    2" finding "for " and "2" respectively).

    .. note::

        This regular expression does not include a match field for the rest of
        the title since Python's regex parser works left-to-right trying to
        find the longest match possible. As a consequence if we added a
        ``(?P<title>.*)`` to the start, we would potentially end up with some
        optional leading parts of a preposition stuck in the title.
    """

    output: MarkdownRecipe
    """
    The parsed markdown recipe structure, populated during rendering and
    returned as the final render result.
    """

    independent_recipe_source_blocks: List[MutableMapping[str, RecipeSourceBlock]]
    """
    Accumulates :py:class:`RecipeSourceBlock` tuples corresponding with recipe
    source blocks in the markdown source.

    The outer list groups the blocks into consecutive series of blocks starting
    with a ``new-recipe`` fenced block (or starting with the first recipe block
    in the file).

    The inner mappings are :py:class:`OrderedDict` objects mapping from
    placeholder strings to :py:class:`RecipeSourceBlock` tuples.
    """

    first_heading: bool
    """
    A flag cleared after the first heading in the source has been rendered.
    """

    def __init__(self, *args: Any, **kwargs: Any) -> None:
        super().__init__(*args, **kwargs)  # type: ignore

        self.output = MarkdownRecipe()
        self.independent_recipe_source_blocks = []
        self.first_heading = True

    def render_heading(self, element: block.Heading) -> str:
        """
        Render headings, also capturing the first heading if it contains a
        serving count.
        """
        level = element.level
        text = self.render_children(element)  # type: ignore

        pre = ""  # To prepend before <h*> tag
        post = ""  # To append after </h*> tag
        attrs = ""  # Attributes to be inserted into the <h*> tag.

        # Capture the first title as recipe title. Ignore it if it isn't a H1
        # level title or if it contains HTML or a scaled value substitution
        # since we can't easily turn these into a plain text title.
        if self.first_heading and level == 1 and "<" not in text and "%" not in text:
            match = self.title_serving_count_pattern.search(text)
            if match is None:
                self.output.title = html.unescape(text.strip())
                attrs = ' class="rg-title-unscalable"'
            else:
                title = text[: match.start()]
                self.output.title = html.unescape(title.strip())
                self.output.servings = int(match["servings"])
                placeholder = generate_placeholder()
                self.output.scaled_value_strings[placeholder] = SVS(
                    int(match["servings"])
                )
                text = title + t(
                    "span",
                    match["preposition"] + placeholder,
                    class_="rg-serving-count",
                )
                attrs = ' class="rg-title-scalable"'
            pre = self.output.pre_title_placeholder = generate_placeholder()
            post = self.output.post_title_placeholder = generate_placeholder()

        self.first_heading = False

        return f"{pre}<h{level}{attrs}>{text}</h{level}>{post}\n"

    def render_scaled_value_expression(self, element: ScaledValueExpression) -> str:
        """Render scaled value expressions."""
        placeholder = generate_placeholder()
        self.output.scaled_value_strings[placeholder] = element.string
        return placeholder

    def render_recipe_source_block(
        self,
        element: Union[CodeBlock, FencedCode],
        in_fenced_block: bool,
    ) -> str:
        """
        Substitute recipe-containing code blocks with placeholders, capturing
        the recipe code.
        """
        # Capture the recipe
        recipe_source_block = RecipeSourceBlock(
            source=element.children[0].children,
            pos=element.pos,
            in_fenced_block=in_fenced_block,
            start_of_new_recipe=element.lang == "new-recipe",
        )

        # Start a new independent recipe if required
        if (
            recipe_source_block.start_of_new_recipe
            or len(self.independent_recipe_source_blocks) == 0
        ):
            self.independent_recipe_source_blocks.append(OrderedDict())

        placeholder = generate_placeholder()
        self.independent_recipe_source_blocks[-1][placeholder] = recipe_source_block

        return placeholder

    def render_code_block(self, element: CodeBlock) -> str:
        """Treat all indented code blocks as a recipe source block."""
        return self.render_recipe_source_block(element, False)

    def render_fenced_code(self, element: FencedCode) -> str:
        """
        Treat fenced code blocks with 'recipe' or 'new-recipe' as recipe source
        blocks. Others are treated as code as usual
        """
        if element.lang in ("recipe", "new-recipe"):
            return self.render_recipe_source_block(element, True)
        else:
            return super().render_fenced_code(element)  # type: ignore

    def render_document(self, element: Document) -> MarkdownRecipe:
        """Render the document and compile all recipe code blocks."""
        # Render the markdown document
        self.output.html = super().render_children(element)  # type: ignore

        # Compile captured recipe blocks
        markdown_source = element.text
        for recipe_source_blocks in self.independent_recipe_source_blocks:
            placeholders = list(recipe_source_blocks.keys())
            sources = [
                recipe_source_block.get_line_number_corrected_source(markdown_source)
                for recipe_source_block in recipe_source_blocks.values()
            ]
            recipes = compile(sources)
            for placeholder, recipe in zip(placeholders, recipes):
                self.output.recipe_placeholders[placeholder] = recipe

        return self.output


class RecipeGrid:
    """
    A :py:mod:`marko` extension which causes the renderer to output a
    :py:class:`MarkdownRecipe` object containing the parsed recipe grid
    markdown document ready to later be scaled and rendered as final HTML.
    """

    elements = [ScaledValueExpression, Document, CodeBlock, FencedCode]
    renderer_mixins = [RecipeGridRendererMixin]


def compile_markdown(markdown_source: str) -> MarkdownRecipe:
    """
    Compile a recipe grid flavoured markdown document, producing a
    :py:class:`MarkdownRecipe` which may be rescaled and rendered into the
    final HTML form as required.

    Internally calls :py:func:`recipe_grid.compiler.compile` and so may throw
    the same kinds of exceptions when syntax errors in the recipe sources are
    encountered.
    """
    return cast(MarkdownRecipe, Markdown(extensions=[RecipeGrid])(markdown_source))



"""
Human-friendly relatively concise number formatting routines.

At a high-level the following function may be used:

.. autofunction:: format_number

Though this in turn is implemented by the following specialised functions

.. autofunction:: format_float

.. autofunction:: format_fraction
"""

import math

from typing import Callable, Union, FrozenSet

from fractions import Fraction


__all__ = [
    "format_float",
    "format_fraction",
    "format_number",
]


def format_float(number: float, significant_figures: int = 3) -> str:
    """
    Format a floating point value in a concise, human-friendly way.

    This formatter will show up to ``significant_figures`` digits after the
    decimal point, with fewer digits being shown for each significant digit in
    the integer part.

    Trailing zeros after the decimal point are dropped, along with the trailing
    decimal point.

    Scientific notation is never used for large values. Significant digits
    before the decimal point are never dropped.
    """
    fractional, integer = math.modf(number)
    integer_str = f"{integer:.0f}"

    integer_digits = len(integer_str.lstrip("-0"))
    fractional_digits = max(0, significant_figures - integer_digits)
    fractional_abs = abs(fractional)
    fractional_str = f"{fractional_abs:.{fractional_digits}f}"[2:].rstrip("0")

    if len(fractional_str) == 0:
        return str(round(number))

    if fractional_str:
        return f"{integer_str}.{fractional_str}"
    else:
        return integer_str


def format_fraction(
    number: Union[int, Fraction],
    allowed_denominators: FrozenSet[int] = frozenset([2, 3, 4, 5, 6, 7, 8, 12, 16]),
    format_float: Callable[[float], str] = format_float,
) -> str:
    """
    Format a :py:class:`~fractions.Fraction` in a human-friendly way.

    Improper fractions are broken down into an integer and fractional part.

    Fractions whose denominator does not appear in ``allowed_denominators`` are
    instead shown as decimal numbers. Formatting of these is deferred to
    the ``format_float`` function which defaults to :py:func:`format_float`.

    Fractions are rendered as ASCII strings in the style '3/4' or, for improper
    fractions, '1 3/4'.
    """
    if number.denominator == 1:  # Integer case
        return str(number.numerator)
    elif number.denominator not in allowed_denominators:  # Decimal fallback case
        return format_float(float(number))
    elif abs(number.numerator) > number.denominator:  # Improper fraction case
        numerator = number.numerator
        denominator = number.denominator

        sign = -1 if numerator < 0 else 1
        numerator = abs(numerator)

        integer_part = numerator // denominator
        numerator %= denominator

        return f"{sign * integer_part} {numerator}/{denominator}"
    else:  # Ordinary fraction case
        return f"{number.numerator}/{number.denominator}"


def format_number(
    number: Union[float, int, Fraction],
    format_float: Callable[[float], str] = format_float,
    format_fraction: Callable[[Fraction], str] = format_fraction,
) -> str:
    """
    Format a number in a human-friendly way.

    If a fraction is given it will be rendered as a fraction, if it is sensible
    to do so. Otherwise, the number will be represented in decimal form.

    See :py:func:`format_float` and :py:func:`format_fraction`.
    """
    if isinstance(number, float):
        return format_float(number)
    else:
        return format_fraction(number)



from typing import Union

import re

from fractions import Fraction


fraction_pattern = re.compile(
    r"((?P<integer>[0-9]+)[ \t]+)?(?P<numerator>[0-9]+)[ \t]*/[ \t]*(?P<denominator>[0-9]+)"
)


def number(value: str) -> Union[int, float, Fraction]:
    """
    Attempt to parse a number formatted as a fraction (e.g. 9 3/4) float (e.g.
    3.14) or integer (e.g. 123). Throws a :py:exc:`ValueError` if this fails.
    """
    match = fraction_pattern.fullmatch(value)
    if match is not None:
        integer = int(match["integer"]) if match["integer"] is not None else 0
        numerator = int(match["numerator"])
        denominator = int(match["denominator"])
        return integer + Fraction(numerator, denominator)
    else:
        try:
            return int(value)
        except ValueError:
            return float(value)



r"""
The :py:mod:`recipe_grid.recipe` module defines the Directed Acyclic Graph
(DAG) data structure used to describe recipes.


Overview
========

Recipes are defined as a series of trees, principally consisting of
:py:class:`Ingredient` nodes and :py:class:`Step` nodes.

An :py:class:`Ingredient` node contains a name and an optional quantity.

A :py:class:`Step` node contains a description of the step and a series of
input nodes (children).

A simple recipe might only consist of a single tree of :py:class:`Ingredient`
and :py:class:`Step` nodes. For example, here's how a simple pasta meal might
be represented.

.. graphviz::

   digraph foo {

      subgraph {
          node [shape=ellipse]

          onion [label=<Onion<br/>(Ingredient)>]
          tomatoes [label=<Chopped Tomatoes<br/>(Ingredient)>]
          pasta [label=<Pasta<br/>(Ingredient)>]
      }

      subgraph {
          node [shape=rectangle]

          chop [label=<Chop<br/>(Step)>]
          fry [label=<Fry<br/>(Step)>]
          boildown [label=<Boil down<br/>(Step)>]
          boil [label=<Boil<br/>(Step)>]
          mix [label=<Mix<br/>(Step)>]
      }

      chop -> onion;

      fry -> chop;

      boildown -> fry;
      boildown -> tomatoes;

      boil -> pasta;

      mix -> boildown;
      mix -> boil;
   }

Sometimes a sub tree within a recipe might be identified as its own sub recipe
and given an emphasising border and title in the rendered recipe table.  For
example, in the recipe above we might make the pasta sauce a sub recipe. To
represent this, a :py:class:`SubRecipe` node type is used which always has exactly one
child and one or more named outputs. In this example, there is only one output
('Pasta sauce').  We'll return to the case with multiple named outputs later.

.. graphviz::

   digraph foo {

      subgraph {
          node [shape=ellipse]

          onion [label=<Onion<br/>(Ingredient)>]
          tomatoes [label=<Chopped Tomatoes<br/>(Ingredient)>]
          pasta [label=<Pasta<br/>(Ingredient)>]
      }

      subgraph {
          node [shape=rectangle]

          chop [label=<Chop<br/>(Step)>]
          fry [label=<Fry<br/>(Step)>]
          boildown [label=<Boil down<br/>(Step)>]
          boil [label=<Boil<br/>(Step)>]
          mix [label=<Mix<br/>(Step)>]
      }

      subgraph {
          node [shape=Mrecord]

          sauce [label="{<fs>Pasta sauce|<fo>(SubRecipe)}"]
      }

      chop -> onion;

      fry -> chop;

      boildown -> fry;
      boildown -> tomatoes;

      sauce:fo -> boildown;

      boil -> pasta;

      mix -> sauce:fs:n;
      mix -> boil;
   }

In other cases, an ingredient or whole sub recipe may be split into parts and
used separately. In this case, a :py:class:`Reference` node may be used to
refer to a :py:class:`SubRecipe` which must be the root of a previous tree in the
recipe. For example, in the recipe below, a Sub Recipe for grated cheese is
referenced in two places, with 75% being mixed into the main recipe and 25%
being used to top it:

.. graphviz::

   digraph foo {

      subgraph {
          node [shape=ellipse]

          cheese [label=<Cheese<br/>(Ingredient)>]
          white_sauce [label=<White sauce<br/>(Ingredient)>]
          pasta [label=<Pasta<br/>(Ingredient)>]
      }

      subgraph {
          node [shape=rectangle]

          grate [label=<Grate<br/>(Step)>]
          mix [label=<Mix<br/>(Step)>]
          top [label=<Top<br/>(Step)>]
      }

      subgraph {
          node [shape=diamond]

          refm [label=<75%<br/>(Reference)>]
          reft [label=<25%<br/>(Reference)>]
      }

      subgraph {
          node [shape=Mrecord]

          grated_cheese [label="{<fc>Grated cheese|<fo>(SubRecipe)}"]
      }

      grate -> cheese;
      grated_cheese:fo -> grate;

      mix -> refm;
      mix -> white_sauce;
      mix -> pasta;

      top -> mix;
      top -> reft;

      subgraph {
            edge [constraint=false,style=dashed];

            refm:sw -> grated_cheese:fc:n
            reft:sw -> grated_cheese:fc:n
      }
   }

A final case is where a :py:class:`SubRecipe` has more than one output. This might occur
when, for example, when vegetables are boiled and both the vegetables and the
water are reserved and used in a later step. Each output may then be referenced
separately by :py:class:`Reference` nodes. :py:class:`SubRecipe` nodes with more than
one output must be the root of a recipe tree.

.. graphviz::

   digraph foo {
      subgraph {
          node [shape=ellipse]

          vegetables [label=<Vegetables<br/>(Ingredient)>]
          gravy_granules [label=<Gravy granules<br/>(Ingredient)>]
      }

      subgraph {
          node [shape=rectangle]

          boil [label=<Boil, reserving water<br/>(Step)>]
          make [label=<Make gravy<br/>(Step)>]
          pour [label=<Pour over<br/>(Step)>]
      }

      subgraph {
          node [shape=diamond]

          refv [label=<(Reference)>]
          refw [label=<(Reference)>]
      }

      subgraph {
          node [shape=Mrecord]

          boiled_veg [label="{{<fv>Boiled Vegetables|<fw>Water}|<fo>(SubRecipe)}"]
      }

      boil -> vegetables;
      boiled_veg -> boil;

      make -> gravy_granules;
      make -> refw;

      pour -> make;
      pour -> refv;

      subgraph {
          edge [constraint=false,style=dashed];

          refv:sw -> boiled_veg:fv:n
          refw:sw -> boiled_veg:fw:n
      }
   }


Data structures
===============

A Recipe is defined at its root by a :py:class:`Recipe` instance. In a simple
document there will be a single :py:class:`Recipe` instance which contains the
entire DAG for that recipe. For more complex documents where the recipe is
presented in several sections, each section of the recipe will be represented
by its own :py:class:`Recipe` with back-references to prior :py:class:`Recipe`\
s in :py:attr:`Recipe.follows`.

.. autoclass:: Recipe
    :members:

DAG Structure
-------------

Recipe trees are in turn defined using subclasses of
:py:class:`RecipeTreeNode`:

.. autoclass:: RecipeTreeNode
    :members:

.. autoclass:: Ingredient
    :members:

.. autoclass:: Step
    :members: description, inputs

.. autoclass:: SubRecipe
    :members: sub_tree, output_names, show_output_names

.. autoclass:: Reference
    :members: output_index, amount


Quantities and Proportions
--------------------------

Absolute quantities (for ingredients and references) and relative proportions
(for references only) are defined as follows:

.. autoclass:: Quantity

.. autoclass:: Proportion

Exceptions
----------

The DAG structure invariants are enforced by checks in their constructors and
any non-conforming input will produce a :py:exc:`RecipeInvariantError`
subclass.

.. autoexception:: RecipeInvariantError

.. autoexception:: MultiOutputSubRecipeUsedAsNonRootNodeError

.. autoexception:: OutputIndexError

.. autoexception:: ZeroOutputSubRecipeError

.. autoexception:: ReferenceToInvalidSubRecipeError

"""


from typing import cast, Union, Optional, Iterable, Tuple, Set

import math

from fractions import Fraction

from dataclasses import dataclass, replace

from recipe_grid.units import UNIT_SYSTEM

from recipe_grid.scaled_value_string import ScaledValueString


class RecipeInvariantError(ValueError):
    """
    Base class for exceptions thrown when an invariant of the
    :py:class:`Recipe` data structure is violated.
    """


class MultiOutputSubRecipeUsedAsNonRootNodeError(RecipeInvariantError):
    """
    Thrown when a :py:class:`SubRecipe` with more than one named output is
    added as non-root node in a recipe tree.
    """


class OutputIndexError(RecipeInvariantError):
    """
    Thrown when a :py:class:`Reference` refers to output which does not exist
    in the referenced :py:class:`SubRecipe`.
    """


class ZeroOutputSubRecipeError(RecipeInvariantError):
    """
    Thrown if a :py:class:`SubRecipe` is defined with no named outputs.
    """


class ReferenceToInvalidSubRecipeError(RecipeInvariantError):
    """
    Thrown if a :py:class:`Reference` refers to a :py:class:`SubRecipe` node
    which is not a root node of a proceeding recipe tree in the same
    :py:class:`Recipe`.
    """


@dataclass(frozen=True)
class Quantity:
    """
    An absolute quantity.

    Suggested rendering::

        q.value
        + (q.value_unit_spacing + q.unit if q.unit is not None else "")
        + q.preposition
    """

    value: Union[int, float, Fraction]
    """
    The quantity of this ingredient to be included. See also the
    :py:attr:`unit` field.
    """

    unit: Optional[str] = None
    """
    The name of the unit of measurement the :py:attr:`amount` field is given
    in. When None, a unit-less quantity is given (e.g. it is a count, e.g. 3
    apples).
    """

    value_unit_spacing: str = ""
    """
    The whitespace to include between the value and unit when displayed, if
    any. If unit is omitted, this should be an empty string.
    """

    preposition: str = ""
    """
    A preposition to append after the value and unit, including any necessary
    leading whitespace. For example " of" for "50g of".
    """

    def has_equal_value_to(self, other: "Quantity") -> bool:
        """
        Compare two quantities, returning true if the quantities are equal (to
        within a small tolerance (allowing for float imprecision) and ignoring
        other metadata.
        """
        other_to_self_scale: Union[int, float, Fraction]
        if self.unit is None and other.unit is None:
            other_to_self_scale = 1
        elif self.unit is None or other.unit is None:
            # Comparison between unit and unitless quantities
            return False
        else:
            try:
                other_to_self_scale = UNIT_SYSTEM.convert_between(
                    other.unit.lower(),
                    self.unit.lower(),
                )
            except KeyError:  # No conversion available
                if self.unit.lower() == other.unit.lower():
                    other_to_self_scale = 1
                else:
                    return False

        return math.isclose(self.value, other.value * other_to_self_scale)

    def scale(self, factor: Union[int, float, Fraction]) -> "Quantity":
        return replace(self, value=self.value * factor)


@dataclass(frozen=True)
class Proportion:
    """
    A relative proportion.

    Suggested rendering::

        (
            (q.value * 100 if q.percentage else q.value)
            if q.value is None else
            q.remainder_wording
        )
        + q.preposition
    """

    value: Optional[Union[int, float, Fraction]] = None
    """
    The proportion (in the range 0.0 to 1.0) of a substance to be used. If
    None, this indicates whatever quantity remains.
    """

    percentage: Optional[bool] = None
    """
    If value is not None, a boolean indicating if the value should be rendered
    as a percentage (i.e. scaled up to the range 0 to 100).

    This field must be None iff value is None.

    If left as None but value is given, this field will be set to False
    automatically.
    """

    remainder_wording: Optional[str] = None
    """
    If value is None, the word to use to indicate "remainder", e.g.
    "remainder", "left over", etc.

    Must be None iff value is not None.

    If left as None but value is also None, this field will be set to
    "remaining" automatically.
    """

    preposition: str = ""
    """
    A preposition to append after the value, including any necessary
    leading whitespace. Examples:

    * "rest of the sauce" -> " of the"
    * "0.5 * sauce" -> " *"
    * "25% of the sauce" -> "% of the"

    Note that for percentages, the preposition string is responsible for
    indicating to a reader that the value is a percentage (e.g. it should
    include a '%' symbol).
    """

    def __post_init__(self) -> None:
        if self.value is not None and self.percentage is None:
            object.__setattr__(self, "percentage", False)
        if self.value is None and self.remainder_wording is None:
            object.__setattr__(self, "remainder_wording", "remaining")

    def scale(self, factor: Union[int, float, Fraction]) -> "Proportion":
        return self  # Nothing to rescale


@dataclass(frozen=True)
class RecipeTreeNode:
    def _assert_can_be_child_node(self) -> None:
        """
        Throws a :py:class:`RecipeInvariantError` if adding this node as a
        child to another node would violate an invariant of the recipe data
        structure.
        """
        pass

    def iter_children(self) -> Iterable["RecipeTreeNode"]:
        """Iterate over the children of this node."""
        return iter(())

    def substitute(
        self, old: "RecipeTreeNode", new: "RecipeTreeNode"
    ) -> "RecipeTreeNode":
        """
        Return a copy of this recipe tree with the node ``old`` replaced with
        ``new``. (The old tree will remain intact).
        """
        if self == old:
            return new
        else:
            return self

    def scale(self, factor: Union[int, float, Fraction]) -> "RecipeTreeNode":
        raise NotImplementedError()


@dataclass(frozen=True)
class Ingredient(RecipeTreeNode):
    """
    A leaf node in a tree describing an ingredient to be used.
    """

    description: ScaledValueString
    """A description of the ingredient."""

    quantity: Optional[Quantity] = None
    """
    The quantity of the ingredient, of specified. If None, the ingredient is
    quantity-less.
    """

    def scale(self, factor: Union[int, float, Fraction]) -> "Ingredient":
        return replace(
            self,
            description=self.description.scale(factor),
            quantity=(
                self.quantity.scale(factor) if self.quantity is not None else None
            ),
        )


@dataclass(frozen=True)
class Step(RecipeTreeNode):
    r"""
    A node in a tree where the node represents a step in a recipe (e.g. 'mix')
    and the children represent the inputs to that step.  Children may be other
    :py:class:`Step` instances, :py:class:`Ingredient` instances or
    :py:class:`Reference` instances referring to outputs of other
    :py:class:`SubRecipe`\ s.
    """

    description: ScaledValueString
    """A description of the step to be carried out."""

    inputs: Tuple[RecipeTreeNode, ...]
    """The inputs to (i.e. children) of this step."""

    def __post_init__(self) -> None:
        for node in self.inputs:
            node._assert_can_be_child_node()

    def iter_children(self) -> Iterable[RecipeTreeNode]:
        return iter(self.inputs)

    def substitute(self, old: RecipeTreeNode, new: RecipeTreeNode) -> RecipeTreeNode:
        if self == old:
            return new
        else:
            return replace(
                self,
                inputs=tuple(node.substitute(old, new) for node in self.inputs),
            )

    def scale(self, factor: Union[int, float, Fraction]) -> "Step":
        return replace(
            self,
            description=self.description.scale(factor),
            inputs=tuple(input.scale(factor) for input in self.inputs),
        )


@dataclass(frozen=True)
class Reference(RecipeTreeNode):
    """
    A reference to a named output of a :py:class:`SubRecipe`.
    """

    sub_recipe: "SubRecipe"
    output_index: int = 0
    """
    The :py:class:`SubRecipe` and output index being referenced. Only sub
    recipes which form the root of a recipe tree may be referenced.
    """

    amount: Union[Quantity, Proportion] = Proportion(1.0)
    """
    The amount of the referenced output to use. Default: all of it.
    """

    def __post_init__(self) -> None:
        if self.output_index >= len(self.sub_recipe.output_names):
            raise OutputIndexError(self.sub_recipe, self.output_index)

    def iter_children(self) -> Iterable[RecipeTreeNode]:
        yield self.sub_recipe

    def substitute(self, old: RecipeTreeNode, new: RecipeTreeNode) -> RecipeTreeNode:
        if self == old:
            return new
        else:
            return replace(
                self,
                sub_recipe=cast(Reference, self.sub_recipe.substitute(old, new)),
            )

    def scale(self, factor: Union[int, float, Fraction]) -> "Reference":
        return replace(
            self,
            sub_recipe=self.sub_recipe.scale(factor),
            amount=self.amount.scale(factor),
        )


@dataclass(frozen=True)
class SubRecipe(RecipeTreeNode):
    """
    A sub recipe is a node representing a logical division in a recipe with
    some semantic significance.  For example, a pie recipe may divide the
    recipe into two sub recipes: one for the filling and another for the
    pastry.
    """

    sub_tree: RecipeTreeNode
    """The steps describing this sub-recipe."""

    output_names: Tuple[ScaledValueString, ...]
    """
    One or more names given to the outputs of this sub recipe.

    In the simple case there will be exactly one named output. In our pie
    recipe example, the sub recipe for the filling might have a single output
    named "Filling" and the pastry sub recipe might have one named "Pastry".

    For sub recipes which produce multiple outputs, names for these must be
    enumerated. For example, a sub recipe describing boiling some vegetables
    where both the vegetables and water will be used, two output names (e.g.
    "Boiled Vegetables" and "Vegetable Water" might be given).

    .. note::

        Sub recipes with a single output name may appear anywhere within a
        recipe tree. Where a SubRecipe is not at the root of a recipe it will
        typically be rendered inline but inset and labelled with the output
        name.

        Sub recipes with more than one named output may only be used as the
        root of a recipe tree.
    """

    show_output_names: bool = True
    """
    Specifies whether the output name(s) for this subrecipe should be rendered.
    For example when a sub-recipe consists of a single ingredient (e.g. '300g
    spam') with a single output (e.g. 'spam'), adding extra labelling would
    just be a distraction and so this setting should be False.
    """

    def __post_init__(self) -> None:
        self.sub_tree._assert_can_be_child_node()

        if len(self.output_names) == 0:
            raise ZeroOutputSubRecipeError()

    def _assert_can_be_child_node(self) -> None:
        if len(self.output_names) > 1:
            raise MultiOutputSubRecipeUsedAsNonRootNodeError()

    def substitute(self, old: RecipeTreeNode, new: RecipeTreeNode) -> RecipeTreeNode:
        if self == old:
            return new
        else:
            return replace(
                self,
                sub_tree=self.sub_tree.substitute(old, new),
            )

    def scale(self, factor: Union[int, float, Fraction]) -> "SubRecipe":
        return replace(
            self,
            sub_tree=self.sub_tree.scale(factor),
            output_names=tuple(
                output_name.scale(factor) for output_name in self.output_names
            ),
        )


@dataclass(frozen=True)
class Recipe:
    """
    A recipe, defined in terms of a series of recipe trees. Later trees may
    reference the outputs of earlier trees resulting in a Directed Acyclic
    Graph (DAG) structure describing a recipe.
    """

    recipe_trees: Tuple[RecipeTreeNode, ...]
    """
    The recipe tree roots for this :py:class:`Recipe`.
    """

    follows: Optional["Recipe"] = None
    r"""
    If this recipe contains :py:class:`Reference`\ s to :py:class:`SubRecipe`\
    s in another :py:class:`Recipe`, this parameter should be set accordingly.
    (References are looked for recursively.
    """

    def __post_init__(self) -> None:
        # Check the consistency of all References (i.e. that they only refer to
        # SubRecipes which appear as roots of recipe trees prior to the tree
        # containing the Reference.
        previous_sub_recipe_roots: Set[SubRecipe] = set()

        next_recipe = self.follows
        while next_recipe is not None:
            previous_sub_recipe_roots.update(
                recipe_tree
                for recipe_tree in next_recipe.recipe_trees
                if isinstance(recipe_tree, SubRecipe)
            )
            next_recipe = next_recipe.follows

        for tree_root in self.recipe_trees:
            to_visit = [tree_root]
            while to_visit:
                node = to_visit.pop()
                if isinstance(node, Reference):
                    if node.sub_recipe not in previous_sub_recipe_roots:
                        raise ReferenceToInvalidSubRecipeError(node)
                to_visit.extend(node.iter_children())

            if isinstance(tree_root, SubRecipe):
                previous_sub_recipe_roots.add(tree_root)

    def scale(self, factor: Union[int, float, Fraction]) -> "Recipe":
        """
        Return a copy of this recipe with all scalable values and quantities
        scaled by the given factor.
        """
        return replace(
            self,
            recipe_trees=tuple(
                recipe_tree.scale(factor) for recipe_tree in self.recipe_trees
            ),
            follows=(self.follows.scale(factor) if self.follows is not None else None),
        )



"""
The ``recipe-grid-site`` command generates recipe book style websites from a
directory of Markdown recipe sources.

Given a directory containing a collection of recipes (as detailed below) the
static site generator is used like so::

    $ recipe-grid-site INPUT_DIR OUTPUT_DIR

Generated websites are entirely static and may be browsed locally or hosted
using any static web hosting service.

Input directory structure
=========================

Each recipe should be placed in its own Markdown file, using a ``.md``
extension. These may be placed in a single directory or organised into a
hierarchy of subdirectories. When subdirectories are used, the generated
website will allow visitors to browse the recipes according to this structure.

An example hierarchy is illustrated below::

    +- my_recipe_website/
    |  +- pasta/
    |  |  +- spag_bol.md
    |  |  +- lasagne.md
    |  |  ...
    |  +- indian/
    |  |  +- mains/
    |  |  |  +- channa_masala.md
    |  |  |  +- tikka_masala.md
    |  |  +- sides/
    |  |  |  +- saag_aloo.md
    |  |  |  +- roti.md
    |  |  ...
    |  ...

Recipe Markdown files must start with a H1-level title giving the name of the
recipe. If this title indicates the number of servings (e.g. 'Tiffin for 2'),
the website will include scaled versions of this recipe to serve between 1 and
10 people (by default). If the number of servings is not indicated, the recipe
will not be scaled and will appear with quantities left as-is.

.. note::

    The maximum serving count shown on the website may be changed from the
    default of 10 using the ``--max-servings`` or ``-s`` argument. The value
    chosen must be at least as large as the largest number of servings any
    recipe is given for.

By default, the recipe website's name and category names are inferred from the
directory names. In the example above, for instance, the website would be named
"My recipe website" and the categories "Pasta", "Indian" and so on. Directory
name are automatically prettified and may be named using ``snake_case``,
``CammelCase`` or ``contain real spaces``.


Readme files
============

You can optionally add a readme files (named ``README.md`` or ``index.md``) to
your directory hierarchy in order customise the website and category names and
to add introductory text.

.. note::

    It doesn't matter if you use ``index.md`` or ``README.md``: both work,
    choose whatever brings you the most fulfilment in life.

The contents of the readme file in the top directory, if provided, will be
shown on the homepage while readme files in subdirectories will be shown on
their respective category browsing pages.

Readme files must start with a H1-level title giving the name for the website
(in the case of the top-level readme) or the name of the category (for other
readme files). These titles override the titles automatically inferred from the
directory names.

These Markdown files are processed as vanilla `CommonMark Markdown
<https://commonmark.org/>`_ and cannot contain recipes themselves.


Links and images
================

All local files and images linked to or used in recipe and readme Markdown
files will be copied into the generated static website.

Links to a recipe markdown file within the site will be rewritten as links to
the compiled recipe page on the website.

Links to subdirectories or readme Markdown files will be rewritten to point to
the corresponding category browsing page.

"""

import sys

from argparse import ArgumentParser

from pathlib import Path

from recipe_grid.static_site.exceptions import StaticSiteError

from recipe_grid.static_site.website import generate_static_site


def main() -> None:
    parser = ArgumentParser(
        description="""
            Compile a directory hierarchy of recipe grid markdown files into a
            static recipe website.
        """,
    )

    parser.add_argument(
        "recipes",
        type=Path,
        help="""
            The directory containing the recipes.
        """,
    )
    parser.add_argument(
        "output",
        type=Path,
        help="""
            The directory to write the generated website to. Will be created if
            it does not exist. Should be empty. Files already in this directory
            may be overwritten silently.
        """,
    )

    parser.add_argument(
        "--max-servings",
        "-s",
        type=int,
        default=10,
        help="""
            The maximum number of servings to scale each recipe to. Must be at
            least as many servings as the largest number of servings originally
            used by any recipe. Default: %(default)s.
        """,
    )

    args = parser.parse_args()

    try:
        generate_static_site(
            args.recipes,
            args.output,
            max_servings=args.max_servings,
        )
    except StaticSiteError as e:
        sys.stderr.write(f"{e}\n")
        sys.exit(1)


if __name__ == "__main__":
    main()



class StaticSiteError(Exception):
    """Base class for exceptions thrown during website generation."""


class NotADirectoryError(StaticSiteError):
    """Thrown when a path given is not a directory."""


class MultipleReadmeError(StaticSiteError):
    """Thrown when a recipe directory contains multiple index.md or readme.md files."""


class RecipeInDirectoryCompileError(StaticSiteError):
    """Thrown when a recipe found in a directory fails to compile."""


class ReadmeMissingTitleError(StaticSiteError):
    """Thrown when an readme.md file is missing a <h1> level title."""


class ReadmeMalformedTitleError(StaticSiteError):
    """Thrown when an readme.md file's <h1> title contains anything but simple text."""


class RecipeMissingTitleError(StaticSiteError):
    """Thrown when an recipe does not have a title."""


class RecipeMissingServingsError(StaticSiteError):
    """Thrown when an recipe does not have a serving count."""


class MaxServingsLowerThanLargestRecipeError(StaticSiteError):
    """Thrown when max_servings set to less than the largest recipes' native serving count."""


class LinkToExternalFileError(StaticSiteError):
    """Thrown when a file outside the source directory is linked to or referenced."""


class LinkToNonExistentFileError(StaticSiteError):
    """Thrown when a non-file is linked to or referenced."""



"""
The ``recipe-grid-lint`` command automatically checks recipes in Markdown files
for common mistakes such as misspelling ingredient or sub recipe names defined
earlier in a recipe.

Usage::

    $ recipe-grid-lint FILENAME [...]

If any potential issues are found, a explanations will be printed to stdout and
a non-zero exit status will be returned. Otherwise, no messages will be
produced and the exit status will be 0.

Errors (e.g. syntax errors) are fatal and must be fixed before this recipe can
be checked any further (or indeed processed by other Recipe Grid tools).

In unusual cases, warnings may be produced for recipes which are actually
correct. If desired, you can suppress warnings of a specific kind using the
``--ignore`` or ``--i`` argument.  Warning types are indicated in square
brackets in warning messages.  This argument may be used multiple times to
ignore multiple kinds of warning.
"""

import sys

from argparse import ArgumentParser

from pathlib import Path

from peggie import ParseError
from recipe_grid.compiler import RecipeCompileError

from recipe_grid.markdown import compile_markdown

from recipe_grid.lint import check, LintKind


def main() -> None:
    parser = ArgumentParser(
        description="""
            Check a recipe grid markdown file for possible mistakes.
        """,
    )

    parser.add_argument(
        "recipe",
        type=Path,
        nargs="*",
        help="""
            The filename of the recipe grid markdown file to compile. Pass
            multiple filenames to check multiple files.
        """,
    )

    parser.add_argument(
        "--ignore",
        "-i",
        action="extend",
        default=[],
        nargs="+",
        choices=[k.name for k in LintKind],
        help="""
            Ignore warnings of a certain types.
        """,
    )

    args = parser.parse_args()

    failed = False
    for page in args.recipe:
        try:
            markdown_recipe = compile_markdown(page.open().read())
            for recipe in markdown_recipe.recipes:
                for lint in check(recipe):
                    if lint.kind.name not in args.ignore:
                        failed = True
                        print(f"{page}: Warning: {lint.description} [{lint.kind.name}]")
        except (ParseError, RecipeCompileError) as e:
            failed = True
            print(f"{page}: Error: {e}")

    sys.exit(1 if failed else 0)


if __name__ == "__main__":
    main()



"""
Recipe Grid has a simple understanding of various cooking-related units and can
perform simple conversions between units of the same kind. The complete list of
units is:

.. rgunitbreakdown::

Conversions are possible betweeen units in the same sublist and names listed
together are aliases for one another.

.. note::

    This system is much less sophisticated than things like :py:mod:`pint` but
    consequently easier to understand and use for non-unit-conversion-related
    tasks and also for less conventional measures (e.g. 'cloves').

Internal API
------------

The Unit system is available in the :py:data:`recipe_grid.units.UNIT_SYSTEM`
which is an instance of the following class:

.. autoclass:: recipe_grid.units.UnitSystem
    :members:
"""

from typing import (
    Optional,
    Union,
    List,
    NamedTuple,
    Set,
    Tuple,
    Mapping,
    MutableMapping,
    Iterable,
    Iterator,
)

from fractions import Fraction

from dataclasses import dataclass, field

import re


__all__ = [
    "Definition",
    "Unit",
    "RelatedUnitSet",
    "UNIT_SYSTEM",
]


Number = Union[int, float, Fraction]


class Definition(NamedTuple):
    quantity: Number
    unit: str


@dataclass(frozen=True)
class Unit:
    """Defines a unit of measurement."""

    names: Tuple[str, ...]
    """The name(s) for this unit."""

    definition: Optional[Definition] = None
    """
    For base units, this is None. For all derived units, a quantity and unit
    defining this unit.
    """

    @property
    def name(self) -> str:
        return self.names[0]

    def __contains__(self, name: str) -> bool:
        """Check if a unit name matches this unit."""
        return name in self.names


@dataclass
class UnitTreeNode:
    """
    For internal use by :py:class:`RelatedUnitSet`. A node in a tree of unit
    definitions.
    """

    name: str

    definition: Optional[Tuple[Number, "UnitTreeNode"]]
    """
    The unit this unit is defined in terms of, along with the quantity of this
    unit amounting to one of the parent unit.
    """

    defines: List["UnitTreeNode"] = field(default_factory=list)
    """
    References to all units defined in terms of this node.
    """


class RelatedUnitSet:
    """A collection of related units."""

    units: List[Unit]

    _name_to_node: MutableMapping[str, UnitTreeNode]

    def __init__(self, units: Iterable[Unit]):
        self.units = list(units)
        self._name_to_node = {}
        for i, unit in enumerate(self.units):
            node: UnitTreeNode
            node_definition: Optional[Tuple[Number, UnitTreeNode]]

            if i == 0:
                if unit.definition is not None:
                    raise ValueError(
                        f"The first Unit, {unit}, must have no definition."
                    )
                node = UnitTreeNode(unit.name, None)
            else:
                if unit.definition is None:
                    raise ValueError(f"Unit, {unit} has no definition.")
                parent_node = self._name_to_node[unit.definition.unit]
                node = UnitTreeNode(unit.name, (unit.definition.quantity, parent_node))
                parent_node.defines.append(node)

            for name in unit.names:
                self._name_to_node[name] = node

    def normalise_unit_name(self, name: str) -> str:
        return self._name_to_node[name].name

    def iter_names(self) -> Iterator[str]:
        return iter(self._name_to_node.keys())

    def __contains__(self, name: str) -> bool:
        """Check if a unit name is defined in this set."""
        return name in self._name_to_node

    def iter_conversions_from(self, from_unit: str) -> Iterator[Tuple[Number, str]]:
        """
        Iterate over all possible conversions from the provided unit to all
        other units in this :py:class:`RelatedUnitSet`.
        """
        visited: Set[str] = set()

        # Queue of (factor, node) tuples
        to_visit: List[Tuple[Number, UnitTreeNode]] = [
            (Fraction(1), self._name_to_node[from_unit])
        ]
        while to_visit:
            scale, node = to_visit.pop(0)

            if node.name in visited:
                continue
            visited.add(node.name)

            yield (scale, node.name)

            if node.definition is not None:
                def_scale, def_node = node.definition
                to_visit.append((scale * def_scale, def_node))

            for child_node in node.defines:
                assert child_node.definition is not None  # Always true when a child
                child_scale = child_node.definition[0]
                to_visit.append((scale / child_scale, child_node))

    def convert_between(self, from_unit: str, to_unit: str) -> Number:
        """
        Find the multiplicative conversion factor to convert from one unit
        to another.
        """
        from_unit = self.normalise_unit_name(from_unit)
        to_unit = self.normalise_unit_name(to_unit)
        for scale, unit in self.iter_conversions_from(from_unit):
            if unit == to_unit:
                return scale
        assert False  # Should be unreachable


@dataclass
class UnitSystem:
    """
    A container of unit conversions.
    """

    unit_sets: Mapping[str, RelatedUnitSet]

    _name_to_unit_set: Mapping[str, RelatedUnitSet] = field(init=False)

    def __post_init__(self) -> None:
        self._name_to_unit_set = {
            name: unit_set
            for unit_set in self.unit_sets.values()
            for name in unit_set.iter_names()
        }

    def __getitem__(self, kind: str) -> RelatedUnitSet:
        return self.unit_sets[kind]

    def __contains__(self, unit_name: str) -> bool:
        """Check if a unit name is supported."""
        return unit_name in self._name_to_unit_set

    def iter_names(self) -> Iterator[str]:
        """Iterate over all defined unit names and aliases."""
        return iter(self._name_to_unit_set.keys())

    def iter_conversions_from(self, from_unit: str) -> Iterator[Tuple[Number, str]]:
        """
        Iterate over all possible conversions from the provided unit to all
        other compatible units. Iterates over pairs of multiplicative
        conversion factor and other unit names.
        """
        return self._name_to_unit_set[from_unit].iter_conversions_from(from_unit)

    def convert_between(self, from_unit: str, to_unit: str) -> Number:
        """
        Find the multiplicative conversion factor to convert from one unit
        to another.
        """
        return self._name_to_unit_set[from_unit].convert_between(from_unit, to_unit)


UNIT_SYSTEM = UnitSystem(
    {
        "mass": RelatedUnitSet(
            [
                # Metric
                Unit(("g", "gram", "grams")),
                Unit(
                    ("kg", "kilo", "kilos", "kilogram", "kilograms"),
                    Definition(1000, "g"),
                ),
                # Crazy
                Unit(("lb", "lbs", "pound", "pounds"), Definition(453.59237, "g")),
                Unit(
                    ("oz", "ozs", "ounce", "ounces"), Definition(Fraction(1, 16), "lb")
                ),
            ]
        ),
        "volume": RelatedUnitSet(
            [
                # Metric
                Unit(("l", "litre")),
                Unit(
                    ("ml", "mill", "mills", "milliliter", "milliliters"),
                    Definition(Fraction(1, 1000), "l"),
                ),
                # Spoons
                Unit(
                    ("tsp", "tsps", "teaspoons", "teaspoon", "tea spoon", "tea spoons"),
                    Definition(Fraction(5, 1000), "l"),
                ),
                Unit(
                    (
                        "tbsp",
                        "tbsps",
                        "tablespoon",
                        "tablespoons",
                        "table spoon",
                        "table spoons",
                    ),
                    Definition(Fraction(15, 1000), "l"),
                ),
                # Crazy
                Unit(("cup", "cups"), Definition(236.58824, "ml")),
                Unit(("pint", "pints"), Definition(568.261, "ml")),
            ]
        ),
        # Other non-specific units
        "clove": RelatedUnitSet([Unit(("clove", "cloves"))]),
        "bulb": RelatedUnitSet([Unit(("bulb", "bulbs"))]),
        "can": RelatedUnitSet([Unit(("can", "cans", "tin", "tins"))]),
        "pinch": RelatedUnitSet([Unit(("pinch", "pinches"))]),
        "knob": RelatedUnitSet([Unit(("knob", "knobs"))]),
        "packet": RelatedUnitSet([Unit(("packet", "packets", "pack", "packs"))]),
        "box": RelatedUnitSet([Unit(("box", "boxes", "boxen"))]),
        "bag": RelatedUnitSet([Unit(("bag", "bags"))]),
        "sack": RelatedUnitSet([Unit(("sack", "sacks"))]),
        "sachet": RelatedUnitSet([Unit(("sachet", "sachets"))]),
        "rasher": RelatedUnitSet([Unit(("rasher", "rashers"))]),
        "strip": RelatedUnitSet([Unit(("strip", "strips"))]),
    }
)
"""
The units known to this system.
"""


ALL_UNITS_REGEX_LITERAL = "|".join(
    re.escape(name).replace(r"\ ", r"\s+") for name in UNIT_SYSTEM.iter_names()
)
"""
A regex literal which matches any of the unit names defined above.
"""



"""
The :py:mod:`peggie` grammar for the recipe specification (read form
``grammar.peg``.

.. autodata:: grammar

.. autodata:: grammar_source

"""

import os

from peggie import compile_grammar, ParseError, RuleExpr, RegexExpr

from recipe_grid.units import ALL_UNITS_REGEX_LITERAL

__all__ = [
    "grammar",
    "grammar_source",
    "prettify_parse_error",
]

grammar_source_path = os.path.join(os.path.dirname(__file__), "grammar.peg")

with open(grammar_source_path) as f:
    grammar_source = f.read().replace("@KNOWN_UNITS@", ALL_UNITS_REGEX_LITERAL)
    """
    The recipe syntax :py:mod:`peggie` grammar source in a string.
    """

grammar = compile_grammar(grammar_source)
"""
The compiled :py:class:`peggie.Grammar` for the recipe syntax.
"""


def prettify_parse_error(parse_error: ParseError) -> ParseError:
    parse_error.expr_explanations = {
        RuleExpr("recipe"): "<action> or <ingredient> or <quantity>",
        RuleExpr("stmt"): "<action> or <ingredient> or <quantity>",
        RuleExpr("ltr_shorthand"): "<ingredient> or <quantity>",
        RuleExpr("expr"): "<action> or <ingredient> or <quantity>",
        RuleExpr("output"): "<output>",
        RuleExpr("action"): "<action>",
        RuleExpr("ingredient"): "<ingredient>",
        RuleExpr("eol"): "<newline>",
        RuleExpr("eof"): "<end of file>",
        RuleExpr("number"): "<number>",
        RuleExpr("interpolated_number"): "<text>",
        RuleExpr("string"): "<text>",
        RuleExpr("static_string"): "<text>",
        RuleExpr("freeform_unit"): "<text>",
        # Add regex descriptions
        RegexExpr("[0-9]+"): "<number>",
        RegexExpr("[^0-9{}\n\r]"): "<text>",
        RegexExpr('[^"\n\r]'): "<text>",
        RegexExpr("[^'\n\r]"): "<text>",
        RegexExpr(":?="): "'=' or ':='",
        # Display literals without escapes
        RegexExpr.literal("("): "'('",
        RegexExpr.literal(")"): "')'",
        RegexExpr.literal(","): "','",
        RegexExpr.literal("{"): "'{'",
        RegexExpr.literal("}"): "'}'",
        RegexExpr.literal("*"): "'*'",
        RegexExpr.literal("%"): "'%'",
        # Omit insignificant whitespace
        RuleExpr("hsp"): None,
        RuleExpr("sp"): None,
    }
    parse_error.last_resort_exprs = {
        RuleExpr("eof"),
        RuleExpr("eol"),
        RegexExpr(r"\\"),
    }
    return parse_error



"""
Abstract Syntax Tree (AST) for the recipe syntax.
"""

from dataclasses import dataclass, field

from fractions import Fraction
from typing import cast, Any, List, Optional, Union, Mapping, Tuple

import peggie

# XXX: Work-around for MyPy issue python/mypy#3186 whereby int and friends are
# not a subtype of numbers.Number.
Number = Union[int, float, Fraction]


@dataclass
class AST:
    """
    Base class for all AST nodes.
    """

    offset: int = field(compare=False)
    """Source offset (in chars) of the corresponding part of the recipe."""


@dataclass
class Recipe(AST):
    """
    Root for all recipe ASTs.
    """

    offset: int = field(init=False, repr=False, compare=False)

    stmts: List["Stmt"]
    """
    A list of :py:class:`Stmt` objects corresponding with the statements in the
    original recipe.
    """

    def __post_init__(self) -> None:
        self.offset = self.stmts[0].offset


@dataclass
class Stmt(AST):
    offset: int = field(init=False, repr=False, compare=False)

    expr: "Expr"
    """The :py:class:`Expr` contained in this statement."""

    outputs: Optional[List["String"]] = None
    """A list of explicitly named outputs produced by this statement."""

    named: bool = False
    """
    True if ``:=`` was used (indicating that this sub-recipe should be
    explicitly named in the displayed output), False otherwise.
    """

    def __post_init__(self) -> None:
        if self.outputs:
            self.offset = self.outputs[0].offset
        else:
            self.offset = self.expr.offset


@dataclass
class Expr(AST):
    """Base class for expression AST nodes."""

    pass


@dataclass
class Step(Expr):
    """A step, e.g. 'mix(tomatoes, herbs)'."""

    offset: int = field(init=False, repr=False, compare=False)

    name: "String"
    """The description of the step ('mix' in this example)."""

    inputs: List["Expr"]
    """The list of inputs to this step, ('tomatoes' and 'herbs' in this example)."""

    def __post_init__(self) -> None:
        self.offset = self.name.offset


@dataclass
class Reference(Expr):
    """A reference to an ingredient or a sub-recipe."""

    offset: int = field(init=False, repr=False, compare=False)

    name: "String"
    """The name of the ingredient or sub-recipe to be used."""

    quantity_or_proportion: Optional[Union["Quantity", "Proportion"]] = None
    """
    The quantity (e.g. '300g') or proportion (e.g. '1/2 *') of the referenced
    ingredient to be included. None if no quantity is specified.
    """

    def __post_init__(self) -> None:
        if self.quantity_or_proportion is not None:
            self.offset = self.quantity_or_proportion.offset
        else:
            self.offset = self.name.offset


@dataclass
class Quantity(AST):
    """An absolute quantity."""

    value: Number
    """
    The scalar quantity. Either a float, or where the amount was given as a
    fraction, a :py:class:`~Fraction`.
    """

    unit: Optional["String"]
    """The name of the unit, or None if a unitless quantity."""

    value_unit_spacing: str
    """The whitespace, if any, used between the value and the unit."""

    preposition: str
    """
    Where a quantity is followed by an unquoted preposition (e.g. the 'of' in
    '50g of butter' or '{1 sack} of pizza'), a string containing that
    preposition. Otherwise an empty string
    """


@dataclass
class Proportion(AST):
    """A relative proportion of a quantity."""

    value: Optional[Number]
    """
    The proportion. A number between 0.0 and 1.0 giving the proportion
    indicated, or None indicating 'rest of'.
    """

    percentage: bool
    """True if the amount was specified as a percentage."""

    remainder_wording: Optional[str]
    """
    When value is None, the string to used to mean 'remainder' (e.g. 'rest' in
    'rest of sauce').
    """

    preposition: str
    """
    The words and whitespace following a value or remainder. Examples:

    * "rest of the spam" -> " of the"
    * "1/3 of sauce" -> " of"
    * "10% spam" -> "%"
    * "0.5 * spam" -> " *"
    """


@dataclass
class String(AST):
    """
    A string which may contain numerical values which will be interpolated as
    the recipe is re-scaled.
    """

    def __init__(
        self,
        value: Union[
            str,
            Tuple[int, str],
            List[Union["Substring", "InterpolatedValue"]],
        ],
    ) -> None:
        if isinstance(value, str):
            self.substrings = [Substring(0, value)]
        elif isinstance(value, tuple):
            self.substrings = [Substring(*value)]
        else:
            self.substrings = value

        self.offset = self.substrings[0].offset

    offset: int = field(init=False, repr=False, compare=False)

    substrings: List[Union["Substring", "InterpolatedValue"]]
    """The components which, when concatenated,  make up the string."""


@dataclass
class Substring(AST):
    """A substring which should be used literally."""

    string: str


@dataclass
class InterpolatedValue(AST):
    """
    A number, which forms part of a string, which should be scaled with the
    recipe.
    """

    number: Number


ESCAPE_CHARS: Mapping[str, str] = {
    "\\": "\\",
    "'": "'",
    '"': '"',
    "a": "\a",
    "b": "\b",
    "f": "\f",
    "n": "\n",
    "r": "\r",
    "t": "\t",
    "v": "\v",
}
"""
An enumeration of all valid escape sequences. All others will be passed through
as a backslash followed by the escaped character.
"""


class RecipeTransformer(peggie.ParseTreeTransformer):
    """
    Transformer which transforms a raw :py:mod:`peggie` parse tree into a more
    friendly :py:class:`AST`.
    """

    def _transform_regex(self, regex: peggie.Regex) -> peggie.Regex:
        return regex

    def decimal(
        self, _pt: peggie.ParseTree, children: Any
    ) -> Tuple[int, Union[int, float]]:
        number = float(children.string)
        if "." not in children.string:
            return children.start, int(number)
        else:
            return children.start, number

    def fraction(self, _pt: peggie.ParseTree, children: Any) -> Tuple[int, Fraction]:
        maybe_integer, numer, _sp1, _slash, _sp2, denom = children
        fraction = Fraction(int(numer.string), int(denom.string))

        offset = numer.start
        integer = 0

        if maybe_integer is not None:
            integer_re, _sp = maybe_integer
            offset = integer_re.start
            integer = int(integer_re.string)

        return offset, integer + fraction

    def naked_string(self, _pt: peggie.ParseTree, children: Any) -> List[Substring]:
        return [Substring(children.start, children.string)]

    def quoted_string(self, _pt: peggie.ParseTree, children: Any) -> List[Substring]:
        q1, body, _q2 = children

        string = "".join(
            char.string
            if isinstance(char, peggie.Regex)
            else ESCAPE_CHARS.get(char[1].string, char[1].string)
            for char in body
        )

        return [Substring(q1.start, string)]

    d_quoted_string = quoted_string
    s_quoted_string = quoted_string

    def bracketed_string(
        self, pt: peggie.ParseTree, children: Any
    ) -> List[Union[Substring, InterpolatedValue]]:
        b1, body, _b2 = children

        out: List[Union[Substring, InterpolatedValue]] = []

        current_string_segment = ""
        current_string_segment_offset = b1.start

        for char in body:
            if isinstance(char, tuple):
                offset, number = char

                if current_string_segment:
                    out.append(
                        Substring(
                            current_string_segment_offset,
                            current_string_segment,
                        )
                    )
                current_string_segment = ""
                current_string_segment_offset = None

                out.append(InterpolatedValue(offset, number))
            else:
                if isinstance(char, peggie.Regex):
                    offset = char.start
                    char = char.string
                else:
                    backslash, char = char
                    offset = backslash.start
                    char = ESCAPE_CHARS.get(char.string, char.string)
                current_string_segment += char
                if current_string_segment_offset is None:
                    current_string_segment_offset = offset

        if current_string_segment_offset is not None:
            out.append(
                Substring(
                    current_string_segment_offset,
                    current_string_segment,
                )
            )

        return out

    def string(self, _pt: peggie.ParseTree, children: Any) -> String:
        first, rest = children

        out = String(first)
        if rest is not None:
            sp, rest_text = rest
            if sp is not None:
                out.substrings.append(Substring(sp.start, sp.string))
            out.substrings.extend(rest_text.substrings)

        return out

    static_string = string

    def known_unit(self, _pt: peggie.ParseTree, children: Any) -> String:
        return String([Substring(children.start, children.string)])

    def proportion(self, pt: peggie.Alt, children: Any) -> Proportion:
        if pt.choice_index == 0:  # Remainder
            remainder, maybe_preposition = children

            preposition = ""
            if maybe_preposition is not None:
                preposition_sp, preposition_word = maybe_preposition
                preposition = preposition_sp.string + preposition_word.string

            return Proportion(
                remainder.start, None, False, remainder.string, preposition
            )
        else:  # Number
            (offset, number), some_preposition = children

            percentage = False
            preposition = ""

            preposition_choice = cast(
                peggie.Alt, cast(peggie.Concat, pt.value).values[1]
            ).choice_index
            if preposition_choice == 0:  # e.g. " of the"
                preposition_sp, preposition_word = some_preposition
                preposition = preposition_sp.string + preposition_word.string
            elif preposition_choice == 1:  # e.g. "% of the" or "%"
                percentage = True
                maybe_sp, perc, maybe_preposition = some_preposition
                if maybe_sp is not None:
                    preposition = maybe_sp.string
                preposition += perc.string
                if maybe_preposition is not None:
                    preposition_sp, preposition_word = maybe_preposition
                    preposition += preposition_sp.string + preposition_word.string
            elif preposition_choice == 2:  # e.g. "*"
                maybe_sp, star = some_preposition
                if maybe_sp is not None:
                    preposition = maybe_sp.string
                preposition += star.string

            if percentage:
                number /= 100

            return Proportion(offset, number, percentage, None, preposition)

    def implicit_quantity(self, _pt: peggie.ParseTree, children: Any) -> Quantity:
        (offset, number), maybe_unit_and_preposition = children

        unit = None
        value_unit_spacing = ""
        preposition = ""
        if maybe_unit_and_preposition is not None:
            maybe_unit_sp, unit, maybe_preposition = maybe_unit_and_preposition
            if maybe_unit_sp is not None:
                value_unit_spacing = maybe_unit_sp.string
            if maybe_preposition is not None:
                preposition_sp, preposition_word = maybe_preposition
                preposition = preposition_sp.string + preposition_word.string

        return Quantity(offset, number, unit, value_unit_spacing, preposition)

    def explicit_quantity(self, pt: peggie.Concat, children: Any) -> Quantity:
        (
            open_bracket,
            _sp1,
            (_offset, number),
            maybe_unit,
            _sp2,
            _close_bracket,
            maybe_preposition,
        ) = children

        unit = None
        value_unit_spacing = ""
        if maybe_unit is not None:
            maybe_unit_sp, unit = maybe_unit
            if maybe_unit_sp is not None:
                value_unit_spacing = maybe_unit_sp.string

        preposition = ""
        if maybe_preposition is not None:
            preposition_sp, preposition_word = maybe_preposition
            preposition = preposition_sp.string + preposition_word.string

        return Quantity(
            open_bracket.start, number, unit, value_unit_spacing, preposition
        )

    def reference(self, _pt: peggie.ParseTree, children: Any) -> Reference:
        maybe_quantity_or_proportion, name = children

        quantity_or_proportion: Optional[Union[Quantity, Proportion]] = None
        if maybe_quantity_or_proportion is not None:
            quantity_or_proportion, _sp = maybe_quantity_or_proportion

        return Reference(name, quantity_or_proportion)

    def step(self, _pt: peggie.ParseTree, children: Any) -> Step:
        (
            name,
            _sp1,
            _open,
            _sp2,
            first_input,
            rest_inputs,
            _comma,
            _sp3,
            _close,
        ) = children

        inputs = [first_input]
        for _sp1, _comma, _sp, next_input in rest_inputs:
            inputs.append(next_input)

        return Step(name, inputs)

    def ltr_shorthand(self, _pt: peggie.ParseTree, children: Any) -> Expr:
        reference, steps = children

        for _sp0, _comma, _sp1, step in steps:
            reference = Step(step, [reference])

        return cast(Expr, reference)

    def expr(self, pt: peggie.Alt, children: Any) -> Expr:
        if pt.choice_index in (0, 1):  # step or reference
            return cast(Expr, children)
        else:  # ltr_shorthand
            _open, _sp1, ltr_shorthand, _sp2, _close = children
            return cast(Expr, ltr_shorthand)

    def output_list(self, _pt: peggie.Alt, children: Any) -> List[String]:
        first, rest = children

        out = [first]
        for _sp1, _comma, _sp1, next in rest:
            out.append(next)

        return out

    def stmt(self, _pt: peggie.Alt, children: Any) -> Stmt:
        maybe_target, expr, _eol = children

        output_list = None
        named = False
        if maybe_target is not None:
            output_list, _sp1, colon, _sp2 = maybe_target
            named = colon.string == ":="

        return Stmt(expr, output_list, named)

    def recipe(self, _pt: peggie.Alt, children: Any) -> Recipe:
        _sp, stmts, _eof = children
        return Recipe(stmts)



r"""
The following output-agnostic data structure is used to represent a recipe in
tabular form.

The basic data structure consists of a :py:class:`Table` object which contains
a set of nested lists containing a 2D array of :py:class:`Cell` and
:py:class:`ExtendedCell` objects. Each entry in the array represents a cell in
the table.

Where a cell spans multiple rows or columns, the top-left cell is defined with
a :py:class:`Cell` instance (with :py:attr:`Cell.rows` and
:py:attr:`Cell.columns` set appropriately) and the occluded cells are filled
with :py:class:`ExtendedCell` instances. For convenience, the
:py:meth:`Table.from_dict` class method is provided which can automatically fill
in :py:class:`ExtendedCell` instances when only :py:class`Cell`\ s are given.

.. autoclass:: Table
    :members:

.. autoclass:: Cell
    :members:
    :undoc-members:

.. autoclass:: ExtendedCell
    :members:
    :undoc-members:

Cells have four borders whose display styles are dictated by the ``border_*``
attributes of the :py:class:`Cell`. The border styles are:

.. autoclass:: BorderType
    :members:
"""

from typing import (
    List,
    Sequence,
    MutableMapping,
    Mapping,
    Tuple,
    Set,
    Optional,
    Union,
    Iterable,
    Iterator,
    Generic,
    TypeVar,
    cast,
)

from dataclasses import dataclass, replace

from enum import Enum, auto


T = TypeVar("T")


class BorderType(Enum):
    none = auto()
    """
    No border. Used only for cells which are to be rendered 'outside' the
    table.
    """

    normal = auto()
    """
    A normal border style which will surround most cells.
    """

    sub_recipe = auto()
    """
    A border drawn round a sub recipe; typically thicker than the normal border
    style.
    """


@dataclass
class Cell(Generic[T]):
    value: T
    """The value contained in this cell."""

    rows: int = 1
    columns: int = 1
    """
    The number of columns (to the right) or rows (below) over which this cell
    extends.
    """

    border_left: BorderType = BorderType.normal
    border_right: BorderType = BorderType.normal
    border_top: BorderType = BorderType.normal
    border_bottom: BorderType = BorderType.normal
    """
    The border styles for this cell.
    """


@dataclass
class ExtendedCell(Generic[T]):
    """
    Represents parts of a table which are filled not with a new cell but with
    the extension of an adjacent cell. (Used as a 'dummy' value in
    :py:class:`Table`.)
    """

    cell: Cell[T]
    """A reference to the cell which occludes this cell."""

    drow: int
    dcolumn: int
    """
    The delta in coordinate from the cell this :py:class:`ExtendedCell`
    resides.
    """


class InconsistentTableLayoutError(ValueError):
    """
    Base class for exceptions thrown when a badly formatted table is provided.
    """


class EmptyTableError(InconsistentTableLayoutError):
    """Thrown when an empty table is given."""


class CellExpectedError(InconsistentTableLayoutError):
    """Thrown when an :py:class:`ExtendedCell` is found when a :py:class:`Cell` was expected."""


class ExtendedCellExpectedError(InconsistentTableLayoutError):
    """Thrown when a :py:class:`Cell` was found and a :py:class:`ExtendedCell` was expected."""


class ExtendedCellReferenceError(InconsistentTableLayoutError):
    """Thrown when an :py:class:`ExtendedCell` has a reference to the wrong cell in it."""


class ExtendedCellCoordinateError(InconsistentTableLayoutError):
    """Thrown when an :py:class:`ExtendedCell` has the wrong coordinates in it."""


class MissingCellError(InconsistentTableLayoutError):
    """Thrown when a missing cell is encountered."""


@dataclass
class Table(Generic[T]):
    cells: Sequence[Sequence[Union[Cell[T], ExtendedCell[T]]]]
    """
    The table cells. A dense 2D array indexed as ``cells[row][column]`` with
    spaces covered by extended cells being denoted using
    :py:class:`ExtendedCell`.
    """

    @classmethod
    def from_dict(cls, table_dict: Mapping[Tuple[int, int], Cell[T]]) -> "Table[T]":
        """
        Construct a :py:class:`Table` from a dictionary mapping (row, column)
        to :py:class:`Cell`.
        """
        # Compute table dimensions
        try:
            rows = max(row + cell.rows for (row, _column), cell in table_dict.items())
            columns = max(
                column + cell.columns for (_row, column), cell in table_dict.items()
            )
        except ValueError:  # Thrown when max is provided an empty list
            raise EmptyTableError()

        # Populate the table
        cells: List[List[Optional[Union[Cell[T], ExtendedCell[T]]]]] = [
            [None for _ in range(columns)] for _ in range(rows)
        ]
        for (row, column), cell in table_dict.items():
            cells[row][column] = cell

            for drow in range(cell.rows):
                for dcolumn in range(cell.columns):
                    if drow != 0 or dcolumn != 0:
                        cells[row + drow][column + dcolumn] = ExtendedCell(
                            cell, drow, dcolumn
                        )

        # Check for missing cells
        for row, row_cells in enumerate(cells):
            for column, maybe_cell in enumerate(row_cells):
                if maybe_cell is None:
                    raise MissingCellError(row, column)

        return cls(cast(Sequence[Sequence[Union[Cell[T], ExtendedCell[T]]]], cells))

    def to_dict(self) -> Mapping[Tuple[int, int], Cell[T]]:
        r"""
        Return a dictionary mapping from (row, column) to :py:class:`Cell`
        (omitting :py:class:`ExtendedCell`\ s).
        """
        return {
            (row, column): cell
            for (row, column), cell in self
            if isinstance(cell, Cell)
        }

    @property
    def columns(self) -> int:
        """Number of columns in this table"""
        return len(self.cells[0])

    @property
    def rows(self) -> int:
        """Number of rows in this table"""
        return len(self.cells)

    def __iter__(
        self,
    ) -> Iterator[Tuple[Tuple[int, int], Union[Cell[T], ExtendedCell[T]]]]:
        """
        Iterate over ((row, column), cell_or_extended_cell) tuples in raster
        scan order.
        """
        for row, row_cells in enumerate(self.cells):
            for column, cell in enumerate(row_cells):
                yield (row, column), cell

    def __getitem__(self, index: Tuple[int, int]) -> Union[Cell[T], ExtendedCell[T]]:
        """Get the entry at the specified row and column."""
        row, column = index
        return self.cells[row][column]

    def __post_init__(self) -> None:
        if len(self.cells) == 0 or len(self.cells[0]) == 0:
            raise EmptyTableError()

        # Verify that the table description is consistent.
        to_check = [
            (row, column) for row in range(self.rows) for column in range(self.columns)
        ]
        present = [(row, column) for (row, column), _cell in self]
        if present != to_check:
            row, column = min(
                (set(to_check) | set(present)) - set(to_check).intersection(present)
            )
            raise MissingCellError(row, column)

        while to_check:
            row, column = to_check.pop(0)

            try:
                cell = self[row, column]
            except IndexError:
                raise MissingCellError(row, column)

            if not isinstance(cell, Cell):
                raise CellExpectedError(cell, row, column)

            # Check that all extended cell objects are consistent
            if cell.columns != 1 or cell.rows != 1:
                for erow in range(row, row + cell.rows):
                    for ecolumn in range(column, column + cell.columns):
                        if erow != row or ecolumn != column:
                            try:
                                extended_cell = self[erow, ecolumn]
                            except IndexError:
                                raise MissingCellError(erow, ecolumn)

                            if not isinstance(extended_cell, ExtendedCell):
                                raise ExtendedCellExpectedError(
                                    extended_cell, erow, ecolumn
                                )

                            if extended_cell.cell != cell:
                                raise ExtendedCellReferenceError(
                                    extended_cell, erow, ecolumn
                                )

                            if (
                                extended_cell.drow != erow - row
                                or extended_cell.dcolumn != ecolumn - column
                            ):
                                raise ExtendedCellCoordinateError(
                                    extended_cell, erow, ecolumn
                                )

                            to_check.remove((erow, ecolumn))


def combine_tables(tables: Iterable[Table[T]], axis: int) -> Table[T]:
    """
    Combine two or more tables by sacking them vertically (axis=0) or
    horizontally (axis=1).
    """
    out: MutableMapping[Tuple[int, int], Cell[T]] = {}

    row_offset = 0
    column_offset = 0
    for table in tables:
        for (row, column), cell in table.to_dict().items():
            out[row + row_offset, column + column_offset] = cell

        if axis == 0:
            row_offset += table.rows
        elif axis == 1:
            column_offset += table.columns

    return Table.from_dict(out)


def right_pad_table(table: Table[T], columns: int) -> Table[T]:
    """
    Expand the provided table to ensure it has at least ``columns`` columns
    wide, extending the width of the right-most cells in each row to make up
    the width.
    """
    if table.columns >= columns:
        return table
    else:
        rightmost_cell_coordinates: Set[Tuple[int, int]] = set()
        for row in range(table.rows):
            last_column = table[row, -1]
            if isinstance(last_column, Cell):
                rightmost_cell_coordinates.add((row, table.columns - 1))
            else:
                rightmost_cell_coordinates.add(
                    (row - last_column.drow, table.columns - last_column.dcolumn - 1)
                )

        return Table.from_dict(
            {
                (row, column): (
                    cell
                    if (row, column) not in rightmost_cell_coordinates
                    else replace(cell, columns=columns - column)
                )
                for (row, column), cell in table.to_dict().items()
            }
        )


def set_border_around_table(table: Table[T], border_type: BorderType) -> Table[T]:
    """
    Return a copy of the provided table with the border style changed around
    all outside cell edges.
    """

    def to_cell_coord(row: int, column: int) -> Tuple[int, int]:
        cell_or_extended_cell = table[row, column]
        if isinstance(cell_or_extended_cell, Cell):
            return (row, column)
        else:
            return (
                row - cell_or_extended_cell.drow,
                column - cell_or_extended_cell.dcolumn,
            )

    left_edge_cells = set(to_cell_coord(row, 0) for row in range(table.rows))
    right_edge_cells = set(
        to_cell_coord(row, table.columns - 1) for row in range(table.rows)
    )
    top_edge_cells = set(to_cell_coord(0, column) for column in range(table.columns))
    bottom_edge_cells = set(
        to_cell_coord(table.rows - 1, column) for column in range(table.columns)
    )

    def cell_with_new_border(row: int, column: int) -> Cell[T]:
        cell = cast(Cell[T], table[row, column])

        changes: MutableMapping[str, BorderType] = {}
        if (row, column) in left_edge_cells:
            changes["border_left"] = border_type
        if (row, column) in right_edge_cells:
            changes["border_right"] = border_type
        if (row, column) in top_edge_cells:
            changes["border_top"] = border_type
        if (row, column) in bottom_edge_cells:
            changes["border_bottom"] = border_type

        if changes:
            return replace(cell, **changes)
        else:
            return cell

    return Table.from_dict(
        {
            (row, column): cell_with_new_border(row, column)
            for (row, column), cell in table
            if isinstance(cell, Cell)
        }
    )



"""
Utilities for enumerating categorised recipes within a directory hierarchy.

Recipes may be arranged in a hierarchy of directories (for example) according
to categories and subcategories. All files with a ``.md`` extension are assumed
to be a recipe grid markdown file describing a single recipe.

Directories may optionally contain an ``index.md`` or ``README.md`` file which
better describes the directory contents. This markdown document must contain a
H1 title at the start which will be used as the user-facing category name. When
no such file is present, the category name will be a prettified version of the
directory name (see :py:func:`dirname_to_title`).
"""

from typing import Tuple, NamedTuple, Optional, List

from pathlib import Path

from functools import lru_cache

import re

import marko  # type: ignore

from html import unescape

from peggie import ParseError
from recipe_grid.compiler import RecipeCompileError

from recipe_grid.markdown import compile_markdown, MarkdownRecipe

from recipe_grid.static_site.exceptions import (
    ReadmeMissingTitleError,
    ReadmeMalformedTitleError,
    RecipeInDirectoryCompileError,
    RecipeMissingTitleError,
    RecipeMissingServingsError,
    MultipleReadmeError,
)


def dirname_to_title(filename: str) -> str:
    """
    Given a filename in snake case, camel case or containing Real Spaces(TM),
    returns a normalised rendering. For example "myDirName", "MY_DIR_NAME" and
    "My dir name" would all become "My dir name".
    """
    # Surround numbers in whitespace
    filename = re.sub(
        r"[0-9]+",
        lambda match: f" {match.group(0)} ",
        filename,
    )

    # Add spaces at camel-case word boundaries
    filename = re.sub(
        r"([^A-Z])([A-Z])",
        lambda match: " ".join(g for g in match.groups() if g is not None),
        filename,
    )

    # Replace all runs of punctuation/spaces into single spaces
    filename = re.sub(r"[^a-zA-Z0-9]+", " ", filename)

    # Strip leading/trailing spaces
    filename = filename.strip()

    # Normalise case
    return " ".join(
        word.title() if i == 0 else word.lower()
        for i, word in enumerate(filename.split())
    )


def compile_readme_markdown(path: Path) -> Tuple[str, str]:
    """
    Read and compile a markdown 'README' document, stripping the <h1> heading
    and returning the title and remainder of the document separately.

    Returns
    =======
    title: str
        The title (free from HTML escape sequences)
    description: str
        The remainder of the compiled markdown HTML source.
    """

    with path.open() as f:
        html = marko.Markdown()(f.read())

    lines = html.splitlines(keepends=True)

    first_line = (lines[0] if lines else "").strip()
    if not (first_line.startswith("<h1>") and first_line.endswith("</h1>")):
        raise ReadmeMissingTitleError(f"{path} must start with a h1-level title.")

    title = first_line[len("<h1>") : -len("</h1>")]
    if "<" in title:
        raise ReadmeMalformedTitleError(
            f"{path} must have only simple text in its  h1 title"
        )

    title = unescape(title)
    description = "".join(lines[1:])

    return title, description


_cached_compile_markdown = lru_cache(compile_markdown)
"""Used in :py:func:`compile_recipe`."""


def compile_recipe_markdown(
    recipe_source: Path, require_title: bool = True, require_servings: bool = True
) -> MarkdownRecipe:
    """
    Compile a recipe from source, converting compilation errors into
    :py:exc:`~RecipeInDirectoryCompileError` and optionally throwing
    :py:exc:`~RecipeMissingTitleError` and
    :py:exc:`~RecipeMissingServingsError` if the title or serving count are not
    given for the recipe.
    """
    with recipe_source.open() as f:
        try:
            # NB: Because we cache based on the markdown contents we will never
            # produce a stale result.
            recipe = _cached_compile_markdown(f.read())
        except (RecipeCompileError, ParseError) as e:
            raise RecipeInDirectoryCompileError(
                f"Error while compiling {recipe_source}: {e}"
            )
    if require_title and recipe.title is None:
        raise RecipeMissingTitleError(f"Recipe {recipe_source} is missing a title")
    if require_servings and recipe.servings is None:
        raise RecipeMissingServingsError(
            f"Recipe {recipe_source} is missing a number of "
            f"servings (e.g. '... for 3' in the title)"
        )

    return recipe


class RecipeDirectoryListing(NamedTuple):
    title: str
    """
    The user-facing title for this directory. Taken from the README if present,
    otherwise generated from the directory name.
    """

    description_html: Optional[str]
    """
    If a README is present, the rendered markdown content, excluding the <h1>
    block.
    """

    description_source: Optional[Path]
    """If a README is present, its filename."""

    subdirectories: List[Path]
    """The subdirectories in this listing, in no particular order."""

    recipes: List[Path]
    """The recipes in this directory in no particular order."""


def enumerate_recipe_directory(directory: Path) -> RecipeDirectoryListing:
    """
    Enumerate the contents of a recipe website directory.
    """
    if not directory.is_dir():
        raise NotADirectoryError(f"{directory} is not a directory")

    description_source: Optional[Path] = None
    subdirectories: List[Path] = []
    recipes: List[Path] = []

    for path in directory.iterdir():
        if path.is_dir():
            subdirectories.append(path)
        elif path.name.lower() in ("readme.md", "index.md"):
            if description_source is None:
                description_source = path
            else:
                raise MultipleReadmeError(
                    f"{directory} contains multiple readme files: "
                    f"{description_source.name} and {path.name}."
                )
        elif path.suffix.lower() == ".md":
            recipes.append(path)

    title: str
    description_html: Optional[str]
    if description_source is None:
        title = dirname_to_title(directory.resolve().name)
        description_html = None
    else:
        title, description_html = compile_readme_markdown(description_source)

    return RecipeDirectoryListing(
        title=title,
        description_html=description_html,
        description_source=description_source,
        subdirectories=subdirectories,
        recipes=recipes,
    )



"""
Routines for post-processing HTML, e.g. rewriting links etc.
"""

from typing import List, Tuple, Optional, Callable, Mapping, MutableMapping

from pathlib import Path

from urllib.parse import urlsplit, urlunsplit, quote, unquote

from copy import deepcopy

import mimetypes

from base64 import b64encode

import lxml.html  # type: ignore

from recipe_grid.static_site.exceptions import (
    LinkToExternalFileError,
    LinkToNonExistentFileError,
)

from recipe_grid.static_site import href


HTMLPostprocessingStage = Callable[[lxml.html.Element], Optional[lxml.html.Element]]
"""
Function which post-processes a lxml.html.Element tree, optionally returning
the modified element (or modifying in place if None is returned).
"""


def postprocess_html(
    html: str,
    complete_document: bool = True,
    stages: List[HTMLPostprocessingStage] = [],
) -> str:
    """
    Post-process a HTML document of fragment with the specified processing
    stages which manipulate the document.

    Parameters
    ==========
    html : str
        The HTML source.
    complete_document : str
        If ``html`` is a complete HTML document, this must be True, otherwise
        for HTML fragments, this must be false.
    stages : [ElementTree -> ElementTree or None, ...]
        A series of post-processing stages which transform the lxml parsed
        representation of the provided HTML. If the function returns None, the
        previous provided element tree will be assumed to be modified in-place.

        Note that when complete_document is False, the provided HTML will be
        logically wrapped in a <div> tag. This tag must not be modified or
        removed. The tag will be dropped again in the final output.
    """
    if complete_document:
        tree = lxml.html.document_fromstring(html)
    else:
        # NB: Wrap in <div> because lxml cannot handle things like bare text or
        # sequences of tags as a fragment.
        tree = lxml.html.fragment_fromstring(f"<div>{html}</div>")

    for stage in stages:
        new_tree = stage(tree)
        if new_tree is not None:
            tree = new_tree

    html = lxml.html.tostring(tree).decode("utf-8")

    # Remove <div> wrapper if fragment was used
    if not complete_document:
        empty, open_div, html = html.partition("<div>")
        assert open_div == "<div>"
        assert empty == ""
        html, close_div, empty = html.rpartition("</div>")
        assert close_div == "</div>"
        assert empty == ""

    return html


def resolve_local_links(
    tree: lxml.etree.Element,
    source: Path,
    root: Path,
    from_path: str,
    source_to_page_paths: Mapping[Path, Tuple[str, bool]],
    filename_to_asset_paths: MutableMapping[Path, str],
    assets_dir_path: str,
) -> None:
    """
    A post-processing stage which resolves local URLs (e.g. to other recipes or
    image files).

    Parameters
    ==========
    tree: lxml.etree.ElementTree
        The tree in which rewrite_links will be called.
    source: Path
        The filename of the source file where the links were originally
        defined.
    root: Path
        The root directory containing the source. References will not be
        allowed to files outside this directory tree.
    from_path: str
        The website page path to the page being transformed. Used to construct
        relative hrefs.
    source_to_page_paths: {source_filename: (site_path, is_scalable), ...}
        A lookup from (resolved) paths to website page paths corresponding
        to that source. The is_scalable boolean indicates if the page is
        scalable (e.g. recipes which can be scaled, or a category page) or not
        (e.g. unscalable recipes). In the former case, the first part of the
        site path may be modified to contain a different serving count.

        Used to substitute links to markdown files with their
        compiled recipe pages and links to directories or READMEs with the
        corresponding website page.

        The paths to recipes should point to the page containing the native
        serving size for that recipe.

        Paths to directories should point to the unscaled version of the
        corresponding page (i.e. "/categories/..." path).

        When substitutions are being made, if we're on a recipe or category
        page scaled to a particular number, links to other recipes will use the
        version of the page with the matching scale. Otherwise, the native
        scaling or unscaled page will be referenced.
    filename_to_asset_paths : {filename: site_path, ...}
        A dictionary mapping from (resolved) filenames to website paths to use
        for that static content. This dictionary will be populated
        automatically with all referenced local files, relocated to a website
        path of "<assets_dir_path>/<original file path relative to root>".
    assets_dir_path : site_path
        The path of the assets directory (*without* a trailing slash) where all
        assets will be copied to.
    """

    def rewrite_link(url: str) -> str:
        parts = urlsplit(url)

        # Don't rewrite external links, or links within the page
        if parts.scheme != "" or parts.netloc != "" or parts.path == "":
            return url

        path = unquote(parts.path)

        # Work out the local file system path the URL points at
        fspath: Path
        if path.startswith("/"):
            fspath = root / Path(*path.split("/")[1:])
        else:
            fspath = source.parent / Path(*path.split("/"))
        fspath = fspath.resolve()

        # Work out where on the website this points
        website_path: str
        if fspath in source_to_page_paths:
            website_path, is_scalable = source_to_page_paths[fspath]

            # Change serving number to match current page, if it's path starts
            # with "/serves", except for unscaled recipes
            if from_path.startswith("/serves") and is_scalable:
                website_path = "/".join(
                    from_path.split("/")[:2] + website_path.split("/")[2:]
                )
        else:
            # Verify that the local file exists and is not outside the source
            # root
            root_parts = root.resolve().parts
            if fspath.parts[: len(root_parts)] != root_parts:
                raise LinkToExternalFileError(
                    f"{source} contains a link to a file outside the website source: {url}"
                )
            if not fspath.is_file():
                raise LinkToNonExistentFileError(
                    f"{source} contains a link to non-existent file: {url}"
                )

            # Add the referenced external file to the assets dict
            website_path = (
                assets_dir_path + "/" + "/".join(fspath.parts[len(root_parts) :])
            )
            filename_to_asset_paths[fspath] = website_path

        # Create a relative link
        relative_path = href.relative(from_path, website_path)

        # Reconstitute the URL
        return urlunsplit(parts._replace(path=quote(relative_path)))

    tree.rewrite_links(rewrite_link)


def add_recipe_scaling_links(
    tree: lxml.etree.Element,
    from_path: str,
    scaled_paths: Mapping[int, str],
    native_servings: int,
) -> None:
    """
    A post-processing stage which adds links to other scalings of a recipe to a
    recipe page.

    Parameters
    ==========
    tree: lxml.etree.ElementTree
        The tree in which rewrite_links will be called.
    from_path: str
        The website page path to the page being transformed. Used to construct
        relative hrefs.
    scaled_paths: {servings: site_path, ...}
        A lookup from serving count to paths to scaled pages.
    native_servings : int
        The native number of servings for this recipe.
    """
    for span in tree.find_class("rg-serving-count"):
        # Generate links with each of the different scalings
        scaling_links = []
        for servings, scaled_path in sorted(scaled_paths.items()):
            span_copy = deepcopy(span)
            number = span_copy.find_class("rg-scaled-value")[0]
            number.text = str(servings)
            link = lxml.html.Element("a", href=href.relative(from_path, scaled_path))
            link.text = span_copy.text
            link.extend(span_copy)
            scaling_links.append(link)

        # Wrap current count in link
        current = lxml.html.Element(
            "a", {"href": "#", "class": "rg-serving-count-current"}
        )
        current.extend(span)
        current.text = span.text
        span.text = ""
        span.append(current)

        # Add dropdown list of scalings (to be revealed by CSS)
        dropdown = lxml.html.Element("ul")
        for link in scaling_links:
            li = lxml.html.Element("li")
            li.append(link)
            dropdown.append(li)
        span.append(dropdown)

    for span in tree.find_class("rg-original-servings"):
        # Wrap original serving count in a link to that number of servings
        link = lxml.html.Element(
            "a", href=href.relative(from_path, scaled_paths[native_servings])
        )
        link.text = span.text
        link.extend(span)
        span.append(link)
        span.text = ""


def embed_local_links_as_data_urls(
    tree: lxml.etree.Element,
    source: Path,
    root: Path,
) -> None:
    """
    A post-processing stage which converts local URLs into data URLs containing
    the linked content.

    Parameters
    ==========
    tree: lxml.etree.ElementTree
        The tree in which rewrite_links will be called.
    source: Path
        The filename of the source file where the links were originally
        defined.
    root: Path
        The root directory containing the source. References will not be
        allowed to files outside this directory tree.
    """

    def rewrite_link(url: str) -> str:
        parts = urlsplit(url)

        # Don't rewrite external links, or links within the page
        if parts.scheme != "" or parts.netloc != "" or parts.path == "":
            return url

        path = unquote(parts.path)

        # Work out the local file system path the URL points at
        fspath: Path
        if path.startswith("/"):
            fspath = root / Path(*path.split("/")[1:])
        else:
            fspath = source.parent / Path(*path.split("/"))
        fspath = fspath.resolve()

        # Verify that the local file exists and is not outside the source
        # root
        root_parts = root.resolve().parts
        if fspath.parts[: len(root_parts)] != root_parts:
            raise LinkToExternalFileError(
                f"{source} contains a link to a file outside the website source: {url}"
            )
        if not fspath.is_file():
            raise LinkToNonExistentFileError(
                f"{source} contains a link to non-existent file: {url}"
            )

        # Guess mimetype
        mimetype, _encoding = mimetypes.guess_type(fspath)
        if mimetype is None:
            mimetype = "application/octet-stream"

        # Base64 encode
        base64_data = b64encode(fspath.open("rb").read()).decode("ascii")

        return f"data:{mimetype};base64,{base64_data}"

    tree.rewrite_links(rewrite_link)



"""
A string type in which embedded numerical values can be re-scaled.

.. autoclass:: ScaledValueString

"""

from typing import Union, Callable, List, Sequence, Tuple, Any

from collections.abc import Iterable

from fractions import Fraction

from recipe_grid.number_formatting import format_number

Number = Union[int, float, Fraction]


__all__ = ["ScaledValueString"]


class ScaledValueString:
    """
    A string which contains some numerical values which may be re-scaled by a
    fixed factor.

    This type is very specifically aimed at strings within recipes of the style
    "divide into 8 burgers about 10cm in diameter". Here when the recipe is
    scaled up or down, the '8' should be scaled but the '10' should not. Here a
    :py:class:`ScaledValueString` would be defined like so:

        >>> for_8_burgers = ScaledValueString(
        ...     ["divide into ", 8, " burgers about 10cm in diameter"]
        ... )

    This could be scaled up like so:

        >>> for_16_burgers = for_8_burgers.scale(2)
        >>> for_16_burgers.render()
        'divide into 16 burgers about 10cm in diameter'
    """

    _string: Tuple[Union[str, Number], ...]

    def __init__(
        self, string: Union[str, Number, Sequence[Union[str, Number]]] = ""
    ) -> None:
        if not isinstance(string, Iterable):
            string = [string]

        # Normalise string by combining adjacent strings
        normalised_string: List[Union[str, Number]] = []
        for part in string:
            if (
                isinstance(part, str)
                and normalised_string
                and isinstance(normalised_string[-1], str)
            ):
                normalised_string[-1] += part
            else:
                normalised_string.append(part)

        self._string = tuple(part for part in normalised_string if part != "")

    def render(
        self,
        format_number: Callable[[Number], str] = format_number,
        format_string: Callable[[str], str] = str,
    ) -> str:
        return "".join(
            format_string(part) if isinstance(part, str) else format_number(part)
            for part in self._string
        )

    def __str__(self) -> str:
        return self.render()

    def __repr__(self) -> str:
        return "{}({!r})".format(
            type(self).__name__,
            (self._string[0] if len(self._string) == 1 else self._string),
        )

    def __add__(
        self, other: Union[str, Number, "ScaledValueString"]
    ) -> "ScaledValueString":
        if not isinstance(other, ScaledValueString):
            other = ScaledValueString(other)
        return ScaledValueString(self._string + other._string)

    def scale(self, multiplier: Number) -> "ScaledValueString":
        return ScaledValueString(
            [
                part if isinstance(part, str) else part * multiplier
                for part in self._string
            ]
        )

    def __hash__(self) -> int:
        return hash(self._string)

    def __eq__(self, other: Any) -> bool:
        return isinstance(other, ScaledValueString) and self._string == other._string

    def lower(self) -> "ScaledValueString":
        return ScaledValueString(
            [part.lower() if isinstance(part, str) else part for part in self._string]
        )

    def upper(self) -> "ScaledValueString":
        return ScaledValueString(
            [part.upper() if isinstance(part, str) else part for part in self._string]
        )

    def lstrip(self) -> "ScaledValueString":
        return ScaledValueString(
            [
                part.lstrip() if isinstance(part, str) and i == 0 else part
                for i, part in enumerate(self._string)
            ]
        )

    def rstrip(self) -> "ScaledValueString":
        return ScaledValueString(
            [
                part.rstrip()
                if isinstance(part, str) and i == len(self._string) - 1
                else part
                for i, part in enumerate(self._string)
            ]
        )

    def strip(self) -> "ScaledValueString":
        return self.lstrip().rstrip()



import os
from jinja2 import Environment, PackageLoader, select_autoescape

env = Environment(
    loader=PackageLoader("recipe_grid", os.path.join("static_site", "templates")),
    autoescape=select_autoescape(["html", "xml"]),
)

homepage_template = env.get_template("homepage.html")
categories_template = env.get_template("categories.html")
recipe_template = env.get_template("recipe.html")
website_css_template = env.get_template("website.css")

standalone_recipe_template = env.get_template("standalone_recipe.html")

tables_only_css_template = env.get_template("recipe_tables.css")



"""
Recipes are rendered from their recipe data model descriptions
(:py:mod:`recipe_grid.recipe`) into tabular form for display.

Table rendering is split into two parts. First an abstract tabular description
is generated and then secondly this is rendered into its final form (i.e.
HTML). This decomposition should make it possible for future non-HTML output
formats to be supported.

The abstract table representation is defined in
:py:mod:`recipe_grid.renderer.table`, the recipe-to-table conversion in
:py:mod:`recipe_grid.renderer.recipe_to_table` and finally, table-to-HTML
conversion in :py:mod:`recipe_grid.renderer.html`

:py:mod:`recipe_grid.renderer.table`: Abstract table description
================================================================

.. automodule:: recipe_grid.renderer.table

:py:mod:`recipe_grid.renderer.recipe_to_table`: :py:class:`~recipe_grid.recipe.Recipe` to :py:class:`~recipe_grid.renderer.table.Table` transformation
======================================================================================================================================================

.. automodule:: recipe_grid.renderer.recipe_to_table

:py:mod:`recipe_grid.renderer.html`: HTML Table Renderer
========================================================

.. automodule:: recipe_grid.renderer.html

"""  # noqa: E501



"""
A collection of (fairly basic) linting functions for sanity checking recipes.

The following function will lint a (series of) :py:class:`Recipe` descriptions.

.. autofunction:: check

Linting errors are described by :py:class:`Lint` objects:

.. autoclass:: Lint
    :members:
    :undoc-members:

Different categories of linting errors are identified by members of the
following enumeration. Further details, however, are only given as
human-readable strings.

.. autoclass:: LintKind
    :members:
    :undoc-members:

"""

from typing import Iterable, Set, List, MutableMapping, Optional

from dataclasses import dataclass

from enum import Enum, auto

from math import isclose

from recipe_grid.units import UNIT_SYSTEM

from recipe_grid.recipe import (
    Recipe,
    SubRecipe,
    RecipeTreeNode,
    Reference,
    Step,
    Ingredient,
    Quantity,
    Proportion,
)


class LintKind(Enum):
    """Kinds of lint."""

    unused_ingredient = auto()
    sub_recipe_quantity_unknown = auto()
    sub_recipe_reference_incompatible_units = auto()
    sub_recipe_reference_non_positive_remainder = auto()
    sub_recipe_not_used_up = auto()
    sub_recipe_used_too_much = auto()


@dataclass(frozen=True)
class Lint:
    """
    A description a piece of lint found in a recipe.
    """

    kind: LintKind
    description: str


def check_for_unused_ingredients(recipe_blocks: Iterable[Recipe]) -> Iterable[Lint]:
    """
    Check for ingredients which are defined but never used.

    For example, in the following recipe::

        1 can of spam
        1 egg, boiled, shelled

        slice(spam, eggs)

    An ingredient 'egg' is defined but then later referenced as 'eggs'. This is
    almost certainly a mistake.

    One way to suppress this warning if you really did intend for an unused
    ingredient to remain listed is to explicitly give it a name, e.g.::

        1 can of spam
        egg = 1 egg, unused

        slice(spam)
    """

    implicit_ingredient_sub_recipes: Set[SubRecipe] = set()
    referenced_sub_recipes: Set[SubRecipe] = set()

    def visit_node(node: RecipeTreeNode) -> None:
        if isinstance(node, Reference):
            referenced_sub_recipes.add(node.sub_recipe)
            # Don't recurse into references
            return
        elif isinstance(node, SubRecipe):
            is_implicit_ingredient_sub_recipe = (
                len(node.output_names) == 1 and node.show_output_names is False
            )
            if is_implicit_ingredient_sub_recipe:
                implicit_ingredient_sub_recipes.add(node)

        for child_node in node.iter_children():
            visit_node(child_node)

    for recipe in recipe_blocks:
        for node in recipe.recipe_trees:
            visit_node(node)

    unused_ingredients = implicit_ingredient_sub_recipes - referenced_sub_recipes

    for unused_ingredient in sorted(
        unused_ingredients, key=lambda sr: str(sr.output_names[0])
    ):
        name = unused_ingredient.output_names[0]
        yield Lint(
            kind=LintKind.unused_ingredient,
            description=f"Ingredient '{name}' was defined but never used.",
        )


def check_sub_recipe_references_sum_to_whole(
    recipe_blocks: Iterable[Recipe],
) -> Iterable[Lint]:
    """
    Check that when a sub recipe used in several places is used completely.

    For example, in the following recipe::

        tomato sauce = 1 can of chopped tomatoes, boiled down

        pour over(
            cook(mix(1/2 of tomato sauce, chicken)),
            1/3 of tomato sauce,
        )

    The tomato sauce is not completely used up.
    """
    # NB: We don't capture (and therefore check) cases where a sub recipe
    # output is not used since this is probably intentional (e.g. a final or
    # discarded component of a recipe).
    #
    # sub_recipe_references[sub_recipe][output_index] = [Reference ...]
    sub_recipe_references: MutableMapping[
        SubRecipe, MutableMapping[int, List[Reference]]
    ] = {}

    def visit_node(node: RecipeTreeNode) -> None:
        if isinstance(node, Reference):
            sub_recipe_references.setdefault(node.sub_recipe, {})
            sub_recipe_references[node.sub_recipe].setdefault(node.output_index, [])
            sub_recipe_references[node.sub_recipe][node.output_index].append(node)
        else:
            for child_node in node.iter_children():
                visit_node(child_node)

    for recipe in recipe_blocks:
        for node in recipe.recipe_trees:
            visit_node(node)

    for sub_recipe, output_references in sub_recipe_references.items():
        for output_index, references in output_references.items():
            output_name = str(sub_recipe.output_names[output_index])
            # If this sub recipe contains a single ingredient, infer the quantity
            node = sub_recipe.sub_tree
            while isinstance(node, Step) and len(node.inputs) == 1:
                node = node.inputs[0]
            total_quantity: Optional[Quantity] = None
            if isinstance(node, Ingredient) and len(sub_recipe.output_names) == 1:
                total_quantity = node.quantity

            # Count up the quantities referred to in the references
            problem_encountered: bool = False
            used_proportion: float = 0.0
            for reference in references:
                if isinstance(reference.amount, Quantity):
                    if total_quantity is None:
                        problem_encountered = True
                        yield Lint(
                            kind=LintKind.sub_recipe_quantity_unknown,
                            description=(
                                f"A quantity "
                                f"({reference.amount.value} {reference.amount.unit}) "
                                f"of {output_name} was referenced but the total "
                                f"amount is not known so cannot be checked."
                            ),
                        )
                    else:
                        try:
                            if (
                                reference.amount.unit is not None
                                and total_quantity.unit is not None
                            ):
                                conversion = UNIT_SYSTEM.convert_between(
                                    reference.amount.unit.lower(),
                                    total_quantity.unit.lower(),
                                )
                            elif (
                                reference.amount.unit is None
                                and total_quantity.unit is None
                            ):
                                conversion = 1.0
                            else:
                                raise KeyError()
                            quantity_used = reference.amount.value * conversion
                            used_proportion += quantity_used / total_quantity.value
                        except KeyError:
                            problem_encountered = True
                            yield Lint(
                                kind=LintKind.sub_recipe_reference_incompatible_units,
                                description=(
                                    f"A reference to sub recipe {output_name} "
                                    f"is given using Incompatible units: "
                                    f"{reference.amount.unit}"
                                ),
                            )
                elif isinstance(reference.amount, Proportion):
                    if reference.amount.value is None:
                        if used_proportion >= 1.0:
                            problem_encountered = True
                            yield Lint(
                                kind=LintKind.sub_recipe_reference_non_positive_remainder,
                                description=(
                                    f"A reference to the remainder of recipe {output_name} "
                                    f"was made while none remains unused."
                                ),
                            )
                        used_proportion = max(1.0, used_proportion)
                    else:
                        used_proportion += reference.amount.value

            if not problem_encountered:
                perc = int(used_proportion * 100)
                if isclose(used_proportion, 1.0, rel_tol=0.02):
                    pass  # Used up (almost) exactly
                elif used_proportion < 1.0:
                    yield Lint(
                        kind=LintKind.sub_recipe_not_used_up,
                        description=(
                            f"Not all of {output_name} was used "
                            f"(about {100 - perc}% remains unused)."
                        ),
                    )
                else:  # used_proportion > 1.0
                    yield Lint(
                        kind=LintKind.sub_recipe_used_too_much,
                        description=(
                            f"More of {output_name} was used than is available "
                            f"(about {perc}% of the total amount used)."
                        ),
                    )


def check(recipe_blocks: Iterable[Recipe]) -> Iterable[Lint]:
    """
    Run all linting checks against a given recipe.
    """
    yield from check_for_unused_ingredients(recipe_blocks)
    yield from check_sub_recipe_references_sum_to_whole(recipe_blocks)



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['recipe_grid',
 'recipe_grid.parser',
 'recipe_grid.renderer',
 'recipe_grid.scripts',
 'recipe_grid.static_site',
 'recipe_grid.static_site.templates']

package_data = \
{'': ['*']}

install_requires = \
['jinja2>=3.1.2,<4.0.0',
 'lxml>=4.6.1,<5.0.0',
 'marko>=0.9.1,<0.10.0',
 'peggie>=0.2.0,<0.3.0']

entry_points = \
{'console_scripts': ['recipe-grid = recipe_grid.scripts.recipe_grid:main',
                     'recipe-grid-lint = '
                     'recipe_grid.scripts.recipe_grid_lint:main',
                     'recipe-grid-site = '
                     'recipe_grid.scripts.recipe_grid_site:main']}

setup_kwargs = {
    'name': 'recipe-grid',
    'version': '2.0.3',
    'description': 'A tool for generating table-based recipe descriptions.',
    'long_description': None,
    'author': 'Jonathan Heathcote',
    'author_email': None,
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://github.com/mossblaser/recipe_grid',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'entry_points': entry_points,
    'python_requires': '>=3.7,<4.0',
}


setup(**setup_kwargs)



"""
The ``recipe-grid`` command compiles a single Markdown recipe into a
stand-alone HTML page.

.. highlight:: bash

Basic usage
===========

.. code:: text

    $ recipe-grid RECIPE_SOURCE [OUTPUT_FILENAME]

This will compile the recipe in the indicated markdown file. If no output
filename is given, the input filename with the suffix replaced with '.html' is
used.

Scaling recipes
===============

You can scale the recipe by an arbitrary factor using the ``--scale`` or ``-S``
argument. This takes integers (e.g. '2'), decimal numbers (e.g. '1.5') and
fractions (e.g. '1/2' or '1 1/3').

When the recipe begins with a H1-level title ending with a number of servings
(e.g. 'Tiffin for 2'), you can alternatively use the ``--servings`` or ``-s``
argument to scale the recipe. This argument takes a a number of servings to
scale the recipe to and computes the scaling factor for you.

Links and images
================

By default links to local files and images are embedded as ``data:`` URLs. This
means that the generated HTML page is completely standalone and does not depend
on any other files. This feature can be disabled using the
``--no-embed-local-links`` or ``-E`` flag.
"""

import sys

from argparse import ArgumentParser

from pathlib import Path

from recipe_grid.number_parser import number

from recipe_grid.static_site.exceptions import StaticSiteError

from recipe_grid.static_site.standalone_page import generate_standalone_page


def main() -> None:
    parser = ArgumentParser(
        description="""
            Compile a recipe grid markdown file into a standalone HTML page.
        """,
    )

    parser.add_argument(
        "recipe",
        type=Path,
        help="""
            The filename of the recipe grid markdown file to compile.
        """,
    )
    parser.add_argument(
        "output",
        type=Path,
        nargs="?",
        default=None,
        help="""
            The output filename for the generated HTML file. Defaults to the
            input filename with the extension replaced with .html if no name is
            given.
        """,
    )

    scaling_group = parser.add_mutually_exclusive_group()

    scaling_group.add_argument(
        "--servings",
        "-s",
        type=int,
        metavar="SERVINGS",
        default=None,
        help="""
            Rescale the recipe to serve the specified number of servings. This
            option requires that the recipe declares the number of servings it
            makes in its title (e.g. the title should end with 'for <N>').
        """,
    )

    scaling_group.add_argument(
        "--scale",
        "-S",
        type=number,
        metavar="MULTIPLIER",
        default=None,
        help="""
            Multiplier to scale the recipe by. May be a decimal (e.g. '3' or
            '3.14') or a fraction (e.g. '1/2' or '9 3/4').
        """,
    )

    parser.add_argument(
        "--embed-local-links",
        "-e",
        action="store_true",
        default=True,
        help="""
            Replace all local link and image URLs with data: URLs embedding the
            linked resource directly into the HTML. This is the default mode.
        """,
    )
    parser.add_argument(
        "--no-embed-local-links",
        "-E",
        action="store_false",
        dest="embed_local_links",
        help="""
            Leave local link and image URLs as they are.
        """,
    )

    args = parser.parse_args()

    try:
        html = generate_standalone_page(
            args.recipe,
            servings=args.servings,
            scale=args.scale,
            embed_local_links=args.embed_local_links,
        )
    except StaticSiteError as e:
        sys.stderr.write(f"{e}\n")
        sys.exit(1)

    output = args.output
    if output is None:
        output = args.recipe.with_suffix(".html")

    with output.open("w") as f:
        f.write(html)


if __name__ == "__main__":
    main()



"""
This module implements :py:class:`~recipe_grid.renderer.table.Table` to HTML
conversion in the following routine:

.. autofunction:: render_recipe_tree

In the generated HTML, the following CSS class names are used which should be
styled accordingly using a suitable stylesheet.

The generated table is mostly self explnatory, though ingredients and
references may contain a ``<ul>`` (with the CSS class
``rg-quantity-conversions``) listing quantities using alternative units. This
list may be styled using CSS as a mouse-over hint (for example) or hidden
entirely, as required.

CSS Classes
===========

In the generated HTML, the following CSS class names are used

* Top level (``<table>``) classes:

  ``rg-table``
      Applied to each generated ``<table>``.

* Cell (``<td>``) semantic classes

  ``rg-ingredient``
      Applied to each cell containing an ingredient.
  ``rg-reference``
      Applied to each cell containing an reference.
  ``rg-step``
      Applied to each cell containing a step description.
  ``rg-sub-recipe-header``
      Applied to each cell acting as a header for a sub recipe.
  ``rg-sub-recipe-outputs``
      Applied to the cell on the right-hand-end of a table representing a sub
      recipe with multiple outputs.

* Cell (``<td>``) border styling classes

  ``rg-border-top-none``, ``rg-border-bottom-none``, ``rg-border-left-none``, ``rg-border-right-none``
      Indicates the respective border should be omitted.
  ``rg-border-top-sub-recipe``, ``rg-border-bottom-sub-recipe``, ``rg-border-left-sub-recipe``, ``rg-border-right-sub-recipe``
      Indicates the respective border should be drawn with an emphasised (e.g.
      bold) stroke.

  Where the above classes are given, they are given for both 'sides' of the
  border (e.g. when ``rg-border-left-none`` is used in one cell, the cell
  immediately to its left always has the matching class
  ``rg-border-right-none``).

  When not overridden by the classes above, all other cell borders should be
  solid (and the ``<table>`` should have no border at all), and inter-cell
  spacing should be collapsed.

* Inline styles

  ``rg-quantity-unitless``
      Applied to the ``<span>`` surrounding the number in a unitless quantity.
      For example the "3" in "3 eggs".
  ``rg-quantity-with-conversions``
     Applied to the ``<span>`` surrounding the number and unit and list of unit
     conversions in a quantity.  For example the "1 tsp <ul>...</ul>" in "1 tsp
     <ul>...</ul> of salt".
  ``rg-quantity-without-conversions``
     Applied to the ``<span>`` surrounding the number and unit in a quantity
     where no unit conversions are given.  For example the "1 sack" in "1 sack
     of potatoes".
  ``rg-quantity-conversions``
      Applied to lists (``<ul>>``) of alternative-unit quantity values.
  ``rg-proportion``
      Applied to the ``<span>`` surrounding the proportion in a reference. For
      example in "1/2 of the sauce", this would surround the "1/2 of the" part.

      .. note::

          The number inside will additionally be enclosed in a ``<span>``
          with the class ``rg-scaled-value``.
  ``rg-proportion-remainder``
      Applied to the ``<span>`` surrounding the proportion indicating a
      remainder in reference. For example in "remainder of the sauce", this
      would surround the "remainder of the" part.
  ``rg-sub-recipe-output-list``
      Applies to the list (``ul``) of output names in a cell containing a
      sub recipe with multiple outputs.
  ``rg-scaled-value``
      Applies to the ``<span>`` wrapping all strings containing scaled values
      in the recipe. For example, ingredient quantities or text wrapped in
      ``{`` and ``}`` in the recipe source.

Example CSS stylesheet:
=======================

A sample stylesheet (actually, the one used by the Recipe Grid static site
generator) is given below:

.. literalinclude:: ../../recipe_grid/static_site/templates/recipe_tables.css
   :language: css


"""  # noqa: E501

from typing import cast, Optional, MutableMapping, List, Union, Tuple

import re

import html

from textwrap import indent

from fractions import Fraction

from xml.sax.saxutils import quoteattr

from recipe_grid.scaled_value_string import ScaledValueString as SVS

from recipe_grid.renderer.table import Table, Cell, BorderType

from recipe_grid.number_formatting import format_number

from recipe_grid.units import UNIT_SYSTEM

from recipe_grid.recipe import (
    RecipeTreeNode,
    Ingredient,
    Reference,
    Step,
    SubRecipe,
    Quantity,
    Proportion,
)

from recipe_grid.renderer.recipe_to_table import recipe_tree_to_table


def t(tag: str, body: Optional[str] = None, **attrs: str) -> str:
    """
    A simple utility function for generating HTML tags.

    Examples::

        >>> t("foo")
        '<foo />'
        >>> t("img", src="file.png")
        '<img src="file.png"/>'
        >>> t("a", "Click here", href="elsewhere.html")
        '<a href="elsewhere.html">Click here</a>'
        >>> t("span", "Hiya", class_="fancy")
        '<span class="fancy">Hiya</span>'
        >>> t("span", "Bye", data__foo="bar")
        '<span data-foo="bar">Hiya</span>'

    Note that trailing underscores (``_``) are trimmed from attribute names and
    double underscores (``__``) are replaced with hyphens.
    """

    attrs_str = " ".join(
        name.rstrip("_").replace("__", "-") + "=" + quoteattr(value)
        for name, value in attrs.items()
    )

    if body is None:
        return f"<{tag} {attrs_str}/>"
    else:
        if "\n" in body:
            body = "\n" + indent(body, "  ").rstrip() + "\n"
        return f"<{tag}{(' ' + attrs_str).rstrip()}>{body}</{tag}>"


def render_number(number: Union[float, int, Fraction]) -> str:
    string = format_number(number)

    match = re.fullmatch(r"((?:\d+ )?)(\d+)/(\d+)", string)
    if match is not None:
        # Format fractions using fancy HTML super/subscripts
        integer, superscript, subscript = match.groups()
        superscript = t("sup", superscript)
        subscript = t("sub", subscript)
        return f"{integer}{superscript}&frasl;{subscript}"
    else:
        return string


def render_quantity(quantity: Quantity) -> str:
    if quantity.unit is None:
        return (
            t(
                "span",
                render_number(quantity.value),
                class_="rg-quantity-unitless rg-scaled-value",
            )
            + html.escape(quantity.preposition)
        )
    else:
        alternative_forms: List[Tuple[Union[float, int, Fraction], str]] = []

        try:
            # Present units ordered such that the current unit goes first
            # followed by all with a non-float conversion and then the
            # remaining scales.
            for scale, name in sorted(
                UNIT_SYSTEM.iter_conversions_from(quantity.unit.lower()),
                key=lambda scale_and_name: (
                    scale_and_name[0] != 1,
                    isinstance(scale_and_name[0], float),
                    scale_and_name[1],
                ),
            ):
                alternative_forms.append((quantity.value * scale, name))

            # Don't use normalised unit name for the 'native' quantity
            assert alternative_forms[0][0] == quantity.value  # Sanity check...
            alternative_forms[0] = (quantity.value, quantity.unit)
        except KeyError:
            # Unknown unit; no conversions available
            alternative_forms.append((quantity.value, quantity.unit))

        all_forms = [
            (
                render_number(value)
                + html.escape(quantity.value_unit_spacing)
                + html.escape(unit)
            )
            for value, unit in alternative_forms
        ]

        if len(all_forms) == 1:
            return (
                t(
                    "span",
                    all_forms[0],
                    class_="rg-quantity-without-conversions rg-scaled-value",
                )
                + html.escape(quantity.preposition)
            )
        else:
            return (
                t(
                    "span",
                    (
                        all_forms[0]
                        + t(
                            "ul",
                            "\n".join(t("li", form) for form in all_forms[1:]),
                            class_="rg-quantity-conversions",
                        )
                    ),
                    class_="rg-quantity-with-conversions rg-scaled-value",
                    tabindex="0",
                )
                + html.escape(quantity.preposition)
            )


def render_proportion(proportion: Proportion) -> str:
    if proportion.value is None:
        return t(
            "span",
            html.escape(
                cast(str, proportion.remainder_wording) + proportion.preposition
            ),
            class_="rg-proportion-remainder",
        )
    else:
        return t(
            "span",
            render_number(
                proportion.value * 100 if proportion.percentage else proportion.value
            )
            + html.escape(proportion.preposition).replace("*", "&times;"),
            class_="rg-proportion",
        )


def render_scaled_value_string(string: SVS) -> str:
    return string.render(
        format_number=lambda number: t(
            "span", render_number(number), class_="rg-scaled-value"
        ),
        format_string=html.escape,
    )


def render_ingredient(ingredient: Ingredient) -> str:
    quantity = (
        render_quantity(ingredient.quantity) + " "
        if ingredient.quantity is not None
        else ""
    )
    description = render_scaled_value_string(ingredient.description)
    return quantity + description


def generate_subrecipe_output_id(
    sub_recipe: SubRecipe, output_index: int, prefix: str
) -> str:
    name = str(sub_recipe.output_names[output_index])
    return prefix + re.sub(r"[^a-zA-Z0-9._-]", "-", name).strip("-")


def render_reference(reference: Reference, id_prefix: str) -> str:
    amount: str
    if isinstance(reference.amount, Quantity):
        amount = render_quantity(reference.amount) + " "
    elif isinstance(reference.amount, Proportion):
        if reference.amount.value != 1.0:
            amount = render_proportion(reference.amount) + " "
        else:
            amount = ""
    else:
        raise NotImplementedError(type(reference.amount))

    name = render_scaled_value_string(
        reference.sub_recipe.output_names[reference.output_index]
    )

    return t(
        "a",
        amount + name,
        href="#"
        + generate_subrecipe_output_id(
            reference.sub_recipe, reference.output_index, id_prefix
        ),
    )


def render_step(step: Step) -> str:
    return render_scaled_value_string(step.description)


def render_sub_recipe_header(sub_recipe: SubRecipe) -> str:
    return render_scaled_value_string(sub_recipe.output_names[0])


def render_sub_recipe_outputs(sub_recipe: SubRecipe, id_prefix: str) -> str:
    return t(
        "ul",
        "\n".join(
            t(
                "li",
                render_scaled_value_string(name),
                id=generate_subrecipe_output_id(sub_recipe, index, id_prefix),
            )
            for index, name in enumerate(sub_recipe.output_names)
        ),
        class_="rg-sub-recipe-output-list",
    )


def render_cell(cell: Cell[RecipeTreeNode], id_prefix: str = "sub-recipe-") -> str:
    spans: MutableMapping[str, str] = {}
    if cell.columns != 1:
        spans["colspan"] = str(cell.columns)
    if cell.rows != 1:
        spans["rowspan"] = str(cell.rows)

    class_names: List[str] = []

    body: str
    if isinstance(cell.value, Ingredient):
        class_names.append("rg-ingredient")
        body = render_ingredient(cell.value)
    elif isinstance(cell.value, Reference):
        class_names.append("rg-reference")
        body = render_reference(cell.value, id_prefix)
    elif isinstance(cell.value, Step):
        class_names.append("rg-step")
        body = render_step(cell.value)
    elif isinstance(cell.value, SubRecipe):
        if len(cell.value.output_names) == 1:
            class_names.append("rg-sub-recipe-header")
            body = render_sub_recipe_header(cell.value)
        else:
            class_names.append("rg-sub-recipe-outputs")
            body = render_sub_recipe_outputs(cell.value, id_prefix)

    for edge in ["left", "right", "top", "bottom"]:
        border_type: BorderType = getattr(cell, f"border_{edge}")
        if border_type != BorderType.normal:
            class_names.append(f"rg-border-{edge}-{border_type.name.replace('_', '-')}")

    return t("td", body, class_=" ".join(class_names), **spans)


def render_table(
    table: Table[RecipeTreeNode],
    id: Optional[str] = None,
    id_prefix: str = "sub-recipe-",
) -> str:
    """
    Renders a recipe grid table as HTML.

    Optionally sets the ``id`` attribute of the generated table to the provided
    string (NB: The ``id_prefix`` is not added to this string.).
    """
    return t(
        "table",
        "\n".join(
            t(
                "tr",
                "\n".join(
                    render_cell(cell, id_prefix)
                    for cell in row_cells
                    if isinstance(cell, Cell)
                ),
            )
            for row_cells in table.cells
        ),
        class_="rg-table",
        **({"id": id} if id is not None else {}),
    )


def render_recipe_tree(
    recipe_tree: RecipeTreeNode, id_prefix: str = "sub-recipe-"
) -> str:
    """
    Render a recipe tree as a HTML table.

    The ``id_prefix`` argument may be used to specify the prefix added to all
    anchor IDs used in this recipe tree. The same prefix must be specified when
    rendering every recipe tree within a series of :py:class:`Recipe` blocks.
    When a single HTML page may contain multiple, separate recipes, different
    prefixes should be used for each to ensure links to not conflict.
    """
    id: Optional[str] = None
    if isinstance(recipe_tree, SubRecipe) and len(recipe_tree.output_names) == 1:
        id = generate_subrecipe_output_id(recipe_tree, 0, id_prefix)

    return render_table(recipe_tree_to_table(recipe_tree), id, id_prefix)



"""
The recipe description language is parsed into an Abstract Syntax Tree
(AST) using :py:func:`recipe_grid.parser.parse`:

.. autofunction:: recipe_grid.parser.parse
"""

from typing import cast

from peggie import Parser, ParseError

from recipe_grid.parser.grammar import grammar, prettify_parse_error

from recipe_grid.parser import ast


def parse(source: str) -> ast.Recipe:
    """
    Parse a recipe into an AST (see :py:mod:`recipe_grid.parser.ast`).

    Raises
    ======
    peggie.ParseError
    """
    parser = Parser(grammar)
    try:
        parse_tree = parser.parse(source)
    except ParseError as e:
        raise prettify_parse_error(e)
    return cast(ast.Recipe, ast.RecipeTransformer().transform(parse_tree))
